import os
from unittest import TestCase
from unittest.mock import Mock, patch

import pandas as pd
import pytest

from sdv.metadata.dataset import Metadata, _load_csv, _parse_dtypes, _read_csv_dtypes


def test__read_csv_dtypes():
    """Test read csv data types"""
    # Run
    table_meta = {
        'fields': {
            'a_field': {
                'type': 'categorical'
            },
            'b_field': {
                'type': 'boolean'
            },
            'c_field': {
                'type': 'datetime',
            },
            'd_field': {
                'type': 'id',
                'subtype': 'string'
            },
            'e_field': {
                'type': 'id',
                'subtype': 'integer'
            }
        }
    }
    result = _read_csv_dtypes(table_meta)

    # Asserts
    assert result == {'d_field': str}


def test__parse_dtypes():
    """Test parse data types"""
    # Run
    data = pd.DataFrame({
        'a_field': ['1996-10-17', '1965-05-23'],
        'b_field': ['7', '14'],
        'c_field': ['1', '2'],
        'd_field': ['other', 'data']
    })
    table_meta = {
        'fields': {
            'a_field': {
                'type': 'datetime',
                'format': '%Y-%m-%d'
            },
            'b_field': {
                'type': 'numerical',
                'subtype': 'integer'
            },
            'c_field': {
                'type': 'id',
                'subtype': 'integer'
            },
            'd_field': {
                'type': 'other'
            }
        }
    }
    result = _parse_dtypes(data, table_meta)

    # Asserts
    expected = pd.DataFrame({
        'a_field': pd.to_datetime(['1996-10-17', '1965-05-23'], format='%Y-%m-%d'),
        'b_field': [7, 14],
        'c_field': [1, 2],
        'd_field': ['other', 'data']
    })
    pd.testing.assert_frame_equal(result, expected)


@patch('sdv.metadata.dataset._parse_dtypes')
@patch('sdv.metadata.dataset.pd.read_csv')
@patch('sdv.metadata.dataset._read_csv_dtypes')
def test__load_csv(rcdtypes_mock, read_csv_mock, pdtypes_mock):
    # Run
    table_meta = {
        'path': 'filename.csv',
        'other': 'stuff'
    }
    result = _load_csv(os.path.join('a', 'path'), table_meta)

    # Asserts
    assert result == pdtypes_mock.return_value
    rcdtypes_mock.assert_called_once_with(table_meta)
    dtypes = rcdtypes_mock.return_value
    read_csv_mock.assert_called_once_with(os.path.join('a', 'path', 'filename.csv'), dtype=dtypes)
    pdtypes_mock.assert_called_once_with(read_csv_mock.return_value, table_meta)


class TestMetadata(TestCase):
    """Test Metadata class."""

    def test__analyze_relationships(self):
        """Test get relationships"""
        # Setup
        metadata = Mock(spec_set=Metadata)
        _metadata = {
            'tables': {
                'test': {
                    'use': True,
                    'name': 'test',
                    'fields': {
                        'test_field': {
                            'ref': {'table': 'table_ref', 'field': 'field_ref'},
                            'name': 'test_field'
                        }
                    }
                },
                'test_not_use': {
                    'use': False,
                    'name': 'test_not_use',
                    'fields': {
                        'test_field_not_use': {
                            'ref': {'table': 'table_ref', 'field': 'field_ref'},
                            'name': 'test_field_not_use'
                        }
                    }
                }
            }
        }
        metadata._metadata = _metadata

        # Run
        Metadata._analyze_relationships(metadata)

        # Asserts
        assert metadata._child_map == {'table_ref': {'test'}}
        assert metadata._parent_map == {'test': {'table_ref'}}

    def test__dict_metadata_list(self):
        """Test dict_metadata"""
        # Run
        metadata = {
            'tables': [
                {
                    'name': 'test',
                    'fields': [
                        {
                            'ref': {
                                'table': 'table_ref',
                                'field': 'field_ref'
                            },
                            'name': 'test_field'
                        }
                    ]
                },
                {
                    'name': 'other',
                    'use': False,
                }
            ]
        }
        result = Metadata._dict_metadata(metadata)

        # Asserts
        expected = {
            'tables': {
                'test': {
                    'fields': {
                        'test_field': {
                            'ref': {
                                'table': 'table_ref',
                                'field': 'field_ref'
                            }
                        }
                    }
                }
            }
        }
        assert result == expected

    def test__dict_metadata_dict(self):
        """Test dict_metadata"""
        # Run
        metadata = {
            'tables': {
                'test': {
                    'fields': {
                        'test_field': {
                            'ref': {
                                'table': 'table_ref',
                                'field': 'field_ref'
                            }
                        }
                    }
                },
                'other': {
                    'use': False,
                }
            }
        }
        result = Metadata._dict_metadata(metadata)

        # Asserts
        expected = {
            'tables': {
                'test': {
                    'fields': {
                        'test_field': {
                            'ref': {
                                'table': 'table_ref',
                                'field': 'field_ref'
                            }
                        }
                    }
                }
            }
        }
        assert result == expected

    @patch('sdv.metadata.dataset.Metadata._analyze_relationships')
    @patch('sdv.metadata.dataset.Metadata._dict_metadata')
    def test___init__default_metadata_dict(self, mock_meta, mock_relationships):
        """Test create Metadata instance default with a dict"""
        # Run
        metadata = Metadata({'some': 'meta'})

        # Asserts
        mock_meta.assert_called_once_with({'some': 'meta'})
        mock_relationships.assert_called_once_with()
        assert metadata.root_path == '.'

    def test_get_children(self):
        """Test get children"""
        # Setup
        metadata = Mock(spec_set=Metadata)
        metadata._child_map = {
            'test': 'child_table'
        }

        # Run
        result = Metadata.get_children(metadata, 'test')

        # Asserts
        assert result == 'child_table'

    def test_get_parents(self):
        """Test get parents"""
        # Setup
        metadata = Mock(spec_set=Metadata)
        metadata._parent_map = {
            'test': 'parent_table'
        }

        # Run
        result = Metadata.get_parents(metadata, 'test')

        # Asserts
        assert result == 'parent_table'

    def test_get_table_meta(self):
        """Test get table meta"""
        # Setup
        metadata = Mock(spec_set=Metadata)
        metadata._metadata = {
            'tables': {
                'test': {'some': 'data'}
            }
        }

        # Run
        result = Metadata.get_table_meta(metadata, 'test')

        # Asserts
        assert result == {'some': 'data'}

    @patch('sdv.metadata.dataset._load_csv')
    def test_load_table(self, mock_load_csv):
        """Test load table"""
        # Setup
        metadata = Mock(spec_set=Metadata)
        metadata.root_path = 'a/path'
        metadata.get_table_meta.return_value = {'some': 'data'}
        mock_load_csv.return_value = 'data'

        # Run
        result = Metadata.load_table(metadata, 'test')

        # Asserts
        assert result == 'data'

        metadata.get_table_meta.assert_called_once_with('test')
        mock_load_csv.assert_called_once_with('a/path', {'some': 'data'})

    def test_get_dtypes_with_ids(self):
        """Test get data types including ids."""
        # Setup
        table_meta = {
            'fields': {
                'item 0': {'type': 'id', 'subtype': 'integer'},
                'item 1': {'type': 'numerical', 'subtype': 'integer'},
                'item 2': {'type': 'numerical', 'subtype': 'float'},
                'item 3': {'type': 'categorical'},
                'item 4': {'type': 'boolean'},
                'item 5': {'type': 'datetime'}
            },
            'primary_key': 'item 0'
        }
        metadata = Mock(spec_set=Metadata)
        metadata.get_table_meta.return_value = table_meta
        metadata._DTYPES = Metadata._DTYPES

        # Run
        result = Metadata.get_dtypes(metadata, 'test', ids=True)

        # Asserts
        expected = {
            'item 0': 'int',
            'item 1': 'int',
            'item 2': 'float',
            'item 3': 'object',
            'item 4': 'bool',
            'item 5': 'datetime64',
        }
        assert result == expected

    def test_get_dtypes_no_ids(self):
        """Test get data types excluding ids."""
        # Setup
        table_meta = {
            'fields': {
                'item 0': {'type': 'id', 'subtype': 'integer'},
                'item 1': {'type': 'numerical', 'subtype': 'integer'},
                'item 2': {'type': 'numerical', 'subtype': 'float'},
                'item 3': {'type': 'categorical'},
                'item 4': {'type': 'boolean'},
                'item 5': {'type': 'datetime'},
            }
        }
        metadata = Mock(spec_set=Metadata)
        metadata.get_table_meta.return_value = table_meta
        metadata._DTYPES = Metadata._DTYPES

        # Run
        result = Metadata.get_dtypes(metadata, 'test')

        # Asserts
        expected = {
            'item 1': 'int',
            'item 2': 'float',
            'item 3': 'object',
            'item 4': 'bool',
            'item 5': 'datetime64',
        }
        assert result == expected

    def test_get_dtypes_error_invalid_type(self):
        """Test get data types with an invalid type."""
        # Setup
        table_meta = {
            'fields': {
                'item': {'type': 'unknown'}
            }
        }
        metadata = Mock(spec_set=Metadata)
        metadata.get_table_meta.return_value = table_meta
        metadata._DTYPES = Metadata._DTYPES

        # Run
        errors = []
        Metadata.get_dtypes(metadata, 'test', errors=errors)

        assert len(errors) == 1

    def test_get_dtypes_error_subtype_numerical(self):
        """Test get data types with an invalid numerical subtype."""
        # Setup
        table_meta = {
            'fields': {
                'item': {'type': 'numerical', 'subtype': 'boolean'}
            }
        }
        metadata = Mock(spec_set=Metadata)
        metadata.get_table_meta.return_value = table_meta
        metadata._DTYPES = Metadata._DTYPES

        # Run
        errors = []
        Metadata.get_dtypes(metadata, 'test', errors=errors)

        assert len(errors) == 1

    def test_get_dtypes_error_subtype_id(self):
        """Test get data types with an invalid id subtype."""
        # Setup
        table_meta = {
            'fields': {
                'item': {'type': 'id', 'subtype': 'boolean'}
            }
        }
        metadata = Mock(spec_set=Metadata)
        metadata.get_table_meta.return_value = table_meta
        metadata._DTYPES = Metadata._DTYPES

        # Run
        errors = []
        Metadata.get_dtypes(metadata, 'test', ids=True, errors=errors)

        assert len(errors) == 1

    def test_load_tables(self):
        """Test get tables"""
        # Setup
        table_names = ['foo', 'bar', 'tar']
        table_data = [
            pd.DataFrame({'foo': [1, 2]}),
            pd.DataFrame({'bar': [3, 4]}),
            pd.DataFrame({'tar': [5, 6]})
        ]
        metadata = Mock(spec_set=Metadata)
        metadata.get_tables.side_effect = table_names
        metadata.load_table.side_effect = table_data

        # Run
        tables = ['table 1', 'table 2', 'table 3']
        result = Metadata.load_tables(metadata, tables=tables)

        # Asserts
        expected = {
            'table 1': pd.DataFrame({'foo': [1, 2]}),
            'table 2': pd.DataFrame({'bar': [3, 4]}),
            'table 3': pd.DataFrame({'tar': [5, 6]})
        }
        assert result.keys() == expected.keys()

        for k, v in result.items():
            pd.testing.assert_frame_equal(v, expected[k])

    def test_get_fields(self):
        """Test get fields"""
        # Setup
        table_meta = {
            'fields': {
                'a_field': 'some data',
                'b_field': 'other data'
            }
        }
        metadata = Mock(spec_set=Metadata)
        metadata.get_table_meta.return_value = table_meta

        # Run
        result = Metadata.get_fields(metadata, 'test')

        # Asserts
        expected = {'a_field': 'some data', 'b_field': 'other data'}
        assert result == expected

        metadata.get_table_meta.assert_called_once_with('test')

    def test_get_primary_key(self):
        """Test get primary key"""
        # Setup
        table_meta = {
            'primary_key': 'a_primary_key'
        }
        metadata = Mock(spec_set=Metadata)
        metadata.get_table_meta.return_value = table_meta

        # Run
        result = Metadata.get_primary_key(metadata, 'test')

        # Asserts
        assert result == 'a_primary_key'
        metadata.get_table_meta.assert_called_once_with('test')

    def test_get_foreign_keys(self):
        """Test get foreign key"""
        # Setup
        metadata = Metadata({
            'tables': {
                'parent': {
                    'fields': {
                        'parent_id': {
                            'type': 'id',
                        }
                    },
                    'primary_key': 'parent_id'
                },
                'child': {
                    'fields': {
                        'parent_id': {
                            'type': 'id',
                            'ref': {
                                'table': 'parent',
                                'field': 'id'
                            }
                        },
                        'parent_id_2': {
                            'type': 'id',
                            'ref': {
                                'table': 'parent',
                                'field': 'id'
                            }
                        },
                    }
                }
            }
        })

        # Run
        result = Metadata.get_foreign_keys(metadata, 'parent', 'child')

        # Asserts
        assert result == ['parent_id', 'parent_id_2']

    def test_add_table_already_exist(self):
        """Try to add a new table that already exist"""
        # Setup
        metadata = Mock(spec_set=Metadata)
        metadata.get_tables.return_value = ['a_table', 'b_table']

        # Run
        with pytest.raises(ValueError):
            Metadata.add_table(metadata, 'a_table')

    def test_add_table_only_name(self):
        """Add table with only the name"""
        # Setup
        metadata = Mock(spec_set=Metadata)
        metadata.get_tables.return_value = ['a_table', 'b_table']
        metadata._metadata = {'tables': dict()}

        # Run
        Metadata.add_table(metadata, 'x_table')

        # Asserts
        expected_table_meta = {
            'fields': dict()
        }

        assert metadata._metadata['tables']['x_table'] == expected_table_meta

        metadata.set_primary_key.call_count == 0
        metadata.add_relationship.call_count == 0

    def test_add_table_with_primary_key(self):
        """Add table with primary key"""
        # Setup
        metadata = Mock(spec_set=Metadata)
        metadata.get_tables.return_value = ['a_table', 'b_table']
        metadata._metadata = {'tables': dict()}

        # Run
        Metadata.add_table(metadata, 'x_table', primary_key='id')

        # Asserts
        expected_table_meta = {
            'fields': dict()
        }

        assert metadata._metadata['tables']['x_table'] == expected_table_meta

        metadata.set_primary_key.assert_called_once_with('x_table', 'id')
        metadata.add_relationship.call_count == 0

    def test_add_table_with_foreign_key(self):
        """Add table with foreign key"""
        # Setup
        metadata = Mock(spec_set=Metadata)
        metadata.get_tables.return_value = ['a_table', 'b_table']
        metadata._metadata = {'tables': dict()}

        # Run
        Metadata.add_table(metadata, 'x_table', parent='users')

        # Asserts
        expected_table_meta = {
            'fields': dict()
        }

        assert metadata._metadata['tables']['x_table'] == expected_table_meta

        metadata.set_primary_key.call_count == 0
        metadata.add_relationship.assert_called_once_with('users', 'x_table', None)

    def test_add_table_with_fields_metadata(self):
        """Add table with fields metadata"""
        # Setup
        metadata = Mock(spec_set=Metadata)
        metadata.get_tables.return_value = ['a_table', 'b_table']
        metadata._metadata = {'tables': dict()}

        # Run
        fields_metadata = {
            'a_field': {'type': 'numerical', 'subtype': 'integer'}
        }

        Metadata.add_table(metadata, 'x_table', fields_metadata=fields_metadata)

        # Asserts
        expected_table_meta = {
            'fields': {
                'a_field': {'type': 'numerical', 'subtype': 'integer'}
            }
        }

        assert metadata._metadata['tables']['x_table'] == expected_table_meta

        metadata.set_primary_key.call_count == 0
        metadata.add_relationship.call_count == 0

    def test_add_table_with_fields_no_data(self):
        """Add table with fields and no data"""
        # Setup
        metadata = Mock(spec_set=Metadata)
        metadata.get_tables.return_value = ['a_table', 'b_table']
        metadata._metadata = {'tables': dict()}

        # Run
        fields = ['a_field', 'b_field']

        Metadata.add_table(metadata, 'x_table', fields=fields)

        # Asserts
        expected_table_meta = {
            'fields': dict()
        }

        assert metadata._metadata['tables']['x_table'] == expected_table_meta

    def test_add_table_with_fields_data(self):
        """Add table with fields and data"""
        # Setup
        metadata = Mock(spec_set=Metadata)
        metadata.get_tables.return_value = ['a_table', 'b_table']
        metadata._metadata = {'tables': dict()}
        metadata._get_field_details.return_value = {
            'a_field': {'type': 'numerical', 'subtype': 'integer'},
            'b_field': {'type': 'boolean'}
        }

        # Run
        fields = ['a_field', 'b_field']
        data = pd.DataFrame({'a_field': [0, 1], 'b_field': [True, False], 'c_field': ['a', 'b']})

        Metadata.add_table(metadata, 'x_table', fields=fields, data=data)

        # Asserts
        expected_table_meta = {
            'fields': {
                'a_field': {'type': 'numerical', 'subtype': 'integer'},
                'b_field': {'type': 'boolean'}
            }
        }

        assert metadata._metadata['tables']['x_table'] == expected_table_meta

        metadata.set_primary_key.call_count == 0
        metadata.add_relationship.call_count == 0

    def test_add_table_with_no_fields_data(self):
        """Add table with data to analyze all"""
        # Setup
        metadata = Mock(spec_set=Metadata)
        metadata.get_tables.return_value = ['a_table', 'b_table']
        metadata._metadata = {'tables': dict()}
        metadata._get_field_details.return_value = {
            'a_field': {'type': 'numerical', 'subtype': 'integer'},
            'b_field': {'type': 'boolean'},
            'c_field': {'type': 'categorical'}
        }

        # Run
        data = pd.DataFrame({'a_field': [0, 1], 'b_field': [True, False], 'c_field': ['a', 'b']})

        Metadata.add_table(metadata, 'x_table', data=data)

        # Asserts
        expected_table_meta = {
            'fields': {
                'a_field': {'type': 'numerical', 'subtype': 'integer'},
                'b_field': {'type': 'boolean'},
                'c_field': {'type': 'categorical'}
            }
        }

        assert metadata._metadata['tables']['x_table'] == expected_table_meta

        metadata.set_primary_key.call_count == 0
        metadata.add_relationship.call_count == 0

    @patch('sdv.metadata.dataset.pd.read_csv')
    def test_add_table_with_data_str(self, mock_read_csv):
        """Add table with data as str"""
        # Setup
        metadata = Mock(spec_set=Metadata)
        metadata.get_tables.return_value = ['a_table', 'b_table']
        metadata._metadata = {'tables': dict()}
        mock_read_csv.return_value = pd.DataFrame({
            'a_field': [0, 1],
            'b_field': [True, False],
            'c_field': ['a', 'b']
        })
        metadata._get_field_details.return_value = {
            'a_field': {'type': 'numerical', 'subtype': 'integer'},
            'b_field': {'type': 'boolean'},
            'c_field': {'type': 'categorical'}
        }

        # Run
        Metadata.add_table(metadata, 'x_table', data='/path/to/file.csv')

        expected_table_meta = {
            'fields': {
                'a_field': {'type': 'numerical', 'subtype': 'integer'},
                'b_field': {'type': 'boolean'},
                'c_field': {'type': 'categorical'}
            },
            'path': '/path/to/file.csv'
        }

        assert metadata._metadata['tables']['x_table'] == expected_table_meta

        metadata.set_primary_key.call_count == 0
        metadata.add_relationship.call_count == 0

    def test_add_table_with_constraints(self):
        """Test the ``Metadata.add_table`` method with constraints.

        Expect that when constraints are provided, the metadata for the
        specified table is created with the given constraints.

        Input:
        - Metadata object
        - Table name of the desired table to add
        - Metadata for the table's fields
        - Constraints for the given table
        Side Effects:
        - An entry is added to the metadata for the provided table, which contains
          the given fields and constraints.
        """
        # Setup
        metadata = Mock(spec_set=Metadata)
        metadata.get_tables.return_value = ['a_table', 'b_table']
        metadata._metadata = {'tables': dict()}

        # Run
        fields_metadata = {
            'a_field': {'type': 'numerical', 'subtype': 'integer'},
            'b_field': {'type': 'numerical', 'subtype': 'integer'}
        }
        constraints = [
            {
                'constraint': 'sdv.constraints.tabular.Inequality',
                'columns': [
                    'a_field',
                    'b_field',
                ],
            }
        ]

        Metadata.add_table(
            metadata,
            'x_table',
            fields_metadata=fields_metadata,
            constraints=constraints,
        )

        # Asserts
        expected_table_meta = {
            'fields': {
                'a_field': {'type': 'numerical', 'subtype': 'integer'},
                'b_field': {'type': 'numerical', 'subtype': 'integer'},
            },
            'constraints': [
                {
                    'constraint': 'sdv.constraints.tabular.Inequality',
                    'columns': [
                        'a_field',
                        'b_field',
                    ],
                },
            ]
        }

        assert metadata._metadata['tables']['x_table'] == expected_table_meta

        metadata.set_primary_key.call_count == 0
        metadata.add_relationship.call_count == 0

    def test_add_relationship_table_no_exist(self):
        """Add relationship table no exist"""
        # Setup
        metadata = Mock(spec_set=Metadata)
        metadata.get_tables.return_value = list()

        # Run
        with pytest.raises(ValueError):
            Metadata.add_relationship(metadata, 'a_table', 'b_table')

    def test_add_relationship_parent_no_exist(self):
        """Add relationship table no exist"""
        # Setup
        metadata = Mock(spec_set=Metadata)
        metadata.get_tables.return_value = ['a_table']

        # Run
        with pytest.raises(ValueError):
            Metadata.add_relationship(metadata, 'a_table', 'b_table')

    def test_add_relationship_already_exist(self):
        """Add relationship already exist"""
        # Setup
        metadata = Mock(spec_set=Metadata)
        metadata.get_tables.return_value = ['a_table', 'b_table']
        metadata.get_parents.return_value = set(['b_table'])

        # Run
        with pytest.raises(ValueError):
            Metadata.add_relationship(metadata, 'a_table', 'b_table')

    def test_add_relationship_parent_no_primary_key(self):
        """Add relationship parent no primary key"""
        # Setup
        metadata = Mock(spec_set=Metadata)
        metadata.get_tables.return_value = ['a_table', 'b_table']
        metadata.get_parents.return_value = set()
        metadata.get_children.return_value = set()
        metadata.get_primary_key.return_value = None

        # Run
        with pytest.raises(ValueError):
            Metadata.add_relationship(metadata, 'a_table', 'b_table')

    def test_set_primary_key(self):
        """Set primary key table no exist"""
        # Setup
        metadata = Mock(spec_set=Metadata)
        metadata.get_tables.return_value = list()
        metadata.get_fields.return_value = {'a_field': {'type': 'id', 'subtype': 'integer'}}
        metadata._metadata = {
            'tables': {
                'a_table': {
                    'fields': {'a_field': {'type': 'id', 'subtype': 'integer'}}
                }
            }
        }

        # Run
        Metadata.set_primary_key(metadata, 'a_table', 'a_field')

        # Asserts
        metadata._check_field.assert_called_once_with('a_table', 'a_field', exists=True)
        metadata.get_fields.assert_called_once_with('a_table')
        metadata._get_key_subtype.assert_called_once_with({'type': 'id', 'subtype': 'integer'})

    def test_add_field(self):
        """Add field table no exist"""
        # Setup
        metadata = Mock(spec_set=Metadata)
        metadata.get_tables.return_value = list()
        metadata._metadata = {
            'tables': {
                'a_table': {'fields': dict()}
            }
        }

        # Run
        Metadata.add_field(metadata, 'a_table', 'a_field', 'id', 'string', None)

        # Asserts
        expected_metadata = {
            'tables': {
                'a_table': {
                    'fields': {'a_field': {'type': 'id', 'subtype': 'string'}}
                }
            }
        }

        assert metadata._metadata == expected_metadata
        metadata._check_field.assert_called_once_with('a_table', 'a_field', exists=False)



import numpy as np
import pandas as pd
from rdt.transformers import FloatFormatter

from sdv.constraints import (
    FixedIncrements, Inequality, Range, ScalarInequality, ScalarRange, create_custom_constraint)
from sdv.sampling.tabular import Condition
from sdv.tabular import GaussianCopula


def _is_valid(col, x):
    return pd.Series([True if x_i > 0 else False for x_i in x['col']])


def _transform(col, x):
    return pd.DataFrame({'col': x['col'] ** 2})


def _reverse_transform(col, x):
    return pd.DataFrame({'col': x['col'] ** .5})


def test_create_custom_constraint(tmpdir):
    """Test the ``create_custom_constraint`` method end to end."""
    # Setup
    custom_constraint = create_custom_constraint(
        _is_valid,
        _transform,
        _reverse_transform
    )('col')

    data = pd.DataFrame({'col': np.random.randint(1, 10, size=100)})
    gc = GaussianCopula(constraints=[custom_constraint])
    gc.fit(data)

    # Run
    sampled = gc.sample(100)

    # Assert
    assert all(sampled > 0)

    # Save
    gc.save(tmpdir / 'test.pkl')

    # Load
    loaded = GaussianCopula.load(tmpdir / 'test.pkl')

    # Sample from loaded
    loaded.sample(100)


def test_create_custom_constraint_transform_is_none(tmpdir):
    """Test the ``create_custom_constraint`` method end to end."""
    # Setup
    custom_constraint = create_custom_constraint(_is_valid)('col')

    data = pd.DataFrame({'col': np.random.randint(1, 10, size=100)})
    gc = GaussianCopula(constraints=[custom_constraint])
    gc.fit(data)

    # Run
    sampled = gc.sample(100)

    # Assert
    assert all(sampled > 0)

    # Save
    gc.save(tmpdir / 'test.pkl')

    # Load
    GaussianCopula.load(tmpdir / 'test.pkl')


def test_invalid_create_custom_constraint():
    """Test the an invalid ``create_custom_constraint`` method end to end.

    It should correctly sample the synthetic data through reject sample.
    """
    # Setup
    custom_constraint = create_custom_constraint(
        lambda _, x: pd.Series([True if x_i > 0 else False for x_i in x['col']]),
        lambda _: pd.DataFrame({'col': [10 / 0] * 100}),
        lambda _, x: pd.DataFrame({'col': x['col'] ** .5})
    )('col')

    data = pd.DataFrame({'col': np.random.randint(1, 10, size=100)})
    gc = GaussianCopula(constraints=[custom_constraint])
    gc.fit(data)

    # Run
    sampled = gc.sample(100)

    # Assert
    assert all(sampled > 0)


def test_FixedIncrements():
    """Test the ``FixedIncrements`` constraint end to end."""
    # Setup
    values = np.random.randint(1, 10, size=20) * 5
    data = pd.DataFrame({'column': values})
    constraint = FixedIncrements(column_name='column', increment_value=5)
    gc = GaussianCopula(constraints=[constraint])
    gc.fit(data)

    # Run
    sampled = gc.sample(100)

    # Assert
    assert all(sampled % 5 == 0)


def test_Inequality():
    """Test the ``Inequality`` constraint end to end."""
    # Setup
    data = pd.DataFrame({
        'low': np.random.randint(1, 10, size=20),
        'high': np.random.randint(10, 20, size=20)
    })
    constraint = Inequality('low', 'high')
    gc = GaussianCopula(constraints=[constraint])
    gc.fit(data)

    # Run
    sampled = gc.sample(10)

    # Assert
    assert all(sampled['low'] <= sampled['high'])


def test_ScalarInequality():
    """Test the ``ScalarInequality`` constraint end to end."""
    # Setup
    data = pd.DataFrame({
        'low': np.random.randint(1, 10, size=20),
    })
    constraint = ScalarInequality(column_name='low', value=11, relation='<')
    gc = GaussianCopula(constraints=[constraint])
    gc.fit(data)

    # Run
    sampled = gc.sample(10)

    # Assert
    assert all(sampled['low'] < 11)


def test_Range():
    """Test the ``Range`` constraint end to end."""
    # Setup
    data = pd.DataFrame({
        'low_column': np.random.randint(1, 5, size=20),
        'middle_column': np.random.randint(6, 10, size=20),
        'high_column': np.random.randint(11, 20, size=20),
    })
    constraint = Range(
        low_column_name='low_column',
        middle_column_name='middle_column',
        high_column_name='high_column',
        strict_boundaries=True
    )

    gc = GaussianCopula(constraints=[constraint])
    gc.fit(data)

    # Run
    sampled = gc.sample(100)

    # Assert
    assert sampled.middle_column.min() >= sampled.low_column.min()
    assert sampled.middle_column.max() <= sampled.high_column.max()


def test_ScalarRange():
    """Test the ``ScalarRange`` constraint end to end."""
    # Setup
    data = pd.DataFrame({
        'column': np.random.randint(6, 10, size=20),
    })
    constraint = ScalarRange(
        column_name='column',
        low_value=5,
        high_value=11,
        strict_boundaries=True
    )

    gc = GaussianCopula(constraints=[constraint])
    gc.fit(data)

    # Run
    sampled = gc.sample(100)

    # Assert
    assert sampled.column.min() >= 5
    assert sampled.column.max() <= 11


def test_ScalarRange_conditions():
    """Test ``ScalarRange`` with conditions.

    This test ensures that the conditions are not altered by the constraint transformation.
    """
    # Setup
    constraint = ScalarRange(column_name='input', low_value=49, high_value=100)
    data = pd.DataFrame({
        'input': [_ for _ in range(50, 80)],
        'output': [np.random.rand() for _ in range(30)]
    })
    condition = Condition({'input': 88}, num_rows=10)
    model = GaussianCopula(
        field_names=['input', 'output'],
        field_transformers={
            'input': FloatFormatter,
            'output': FloatFormatter
        },
        constraints=[constraint],
    )

    # Run
    model.fit(data)
    sampled = model.sample_conditions([condition])

    # Assert
    assert all(sampled['input'] == 88)
    assert all(sampled['output'] > 0)
    assert all(sampled['output'] < 1)



"""Tests for the sdv.constraints.base module."""
import re
from unittest.mock import Mock, patch

import pandas as pd
import pytest
from copulas.univariate import GaussianUnivariate

from sdv.constraints.base import (
    ColumnsModel, Constraint, _get_qualified_name, _module_contains_callable_name, get_subclasses,
    import_object)
from sdv.constraints.errors import MissingConstraintColumnError
from sdv.constraints.tabular import FixedCombinations
from sdv.errors import ConstraintsNotMetError


def test__get_qualified_name_class():
    """Test the ``_get_qualified_name`` function, if a class is passed.

    The ``_get_qualified_name`` function is expected to:
    - Return the Fully Qualified Name from a class.

    Input:
    - A class.
    Output:
    - The class qualified name.
    """
    # Run
    fully_qualified_name = _get_qualified_name(Constraint)

    # Assert
    expected_name = 'sdv.constraints.base.Constraint'
    assert fully_qualified_name == expected_name


def test__get_qualified_name_class_has_no_name():
    """Test the ``_get_qualified_name`` function, if a class is passed but no name on it.

    The ``_get_qualified_name`` function is expected to:
    - Return the Fully Qualified Name from a class.

    Input:
    - A class.
    Output:
    - The class qualified name.
    """
    # Run
    fully_qualified_name = _get_qualified_name(Mock())

    # Assert
    expected_name = 'unittest.mock.Mock'
    assert fully_qualified_name == expected_name


def test__get_qualified_name_function():
    """Test the ``_get_qualified_name`` function, if a function is passed.

    The ``_get_qualified_name`` function is expected to:
    - Return the Fully Qualified Name from a function.

    Input:
    - A function.
    Output:
    - The function qualified name.
    """
    # Run
    fully_qualified_name = _get_qualified_name(_get_qualified_name)

    # Assert
    expected_name = 'sdv.constraints.base._get_qualified_name'
    assert fully_qualified_name == expected_name


def test__module_contains_callable_name():
    """Test the ``_module_constraint_callable_name``.

    Return whether or not the ``module`` contains a callable name from this object.
    """
    # Setup
    result = _module_contains_callable_name(Mock())

    # Assert
    assert result


def test_get_subclasses():
    """Test the ``get_subclasses`` function.

    The ``get_subclasses`` function is expected to:
    - Recursively find subclasses for the class object passed.

    Setup:
    - Create three classes, Parent, Child and GrandChild,
      which inherit of each other hierarchically.

    Input:
    - The Parent class.
    Output:
    - Dict of the subclasses of the class: ``Child`` and ``GrandChild`` classes.
    """
    # Setup
    class Parent:
        pass

    class Child(Parent):
        pass

    class GrandChild(Child):
        pass

    # Run
    subclasses = get_subclasses(Parent)

    # Assert
    expected_subclasses = {
        'Child': Child,
        'GrandChild': GrandChild
    }

    assert subclasses == expected_subclasses


def test_import_object_class():
    """Test the ``import_object`` function, when importing a class.

    The ``import_object`` function is expected to:
    - Import a class from its qualifed name.

    Input:
    - Qualified name of the class.
    Output:
    - The imported class.
    """
    # Run
    obj = import_object('sdv.constraints.base.Constraint')

    # Assert
    assert obj is Constraint


def test_import_object_function():
    """Test the ``import_object`` function, when importing a function.

    The ``import_object`` function is expected to:
    - Import a function from its qualifed name.

    Input:
    - Qualified name of the function.
    Output:
    - The imported function.
    """
    # Run
    imported = import_object('sdv.constraints.base.import_object')

    # Assert
    assert imported is import_object


class TestConstraint():

    def test_fit(self):
        """Test the ``Constraint.fit`` method.

        The base ``Constraint.fit`` method is expected to:
        - Call ``_fit`` method.

        Input:
        - Table data (pandas.DataFrame)
        """
        # Setup
        table_data = pd.DataFrame({
            'a': [1, 2, 3]
        })
        instance = Constraint()
        instance._fit = Mock()
        instance._validate_data_meets_constraint = Mock()

        # Run
        instance.fit(table_data)

        # Assert
        instance._fit.assert_called_once_with(table_data)
        instance._validate_data_meets_constraint.assert_called_once_with(table_data)

    def test__validate_data_meets_constraints(self):
        """Test the ``_validate_data_meets_constraint`` method.

        Expect that the method calls ``is_valid`` when the constraint columns
        are in the given data.

        Input:
        - Table data
        Output:
        - None
        Side Effects:
        - No error
        """
        # Setup
        data = pd.DataFrame({
            'a': [0, 1, 2],
            'b': [3, 4, 5]
        }, index=[0, 1, 2])
        constraint = Constraint()
        constraint.constraint_columns = ['a', 'b']
        constraint.is_valid = Mock()

        # Run
        constraint._validate_data_meets_constraint(data)

        # Assert
        constraint.is_valid.assert_called_once_with(data)

    def test__validate_data_meets_constraints_invalid_input(self):
        """Test the ``_validate_data_meets_constraint`` method.

        Expect that the method raises an error when the constraint columns
        are in the given data and the ``is_valid`` returns False for any row.

        Input:
        - Table data contains an invalid row
        Output:
        - None
        Side Effects:
        - A ``ConstraintsNotMetError`` is thrown
        """
        # Setup
        data = pd.DataFrame({
            'a': [0, 1, 2, 3, 4, 5, 6, 7],
            'b': [3, 4, 5, 6, 7, 8, 9, 10]
        }, index=[0, 1, 2, 3, 4, 5, 6, 7])
        constraint = Constraint()
        constraint.constraint_columns = ['a', 'b']
        is_valid_result = pd.Series([True, False, True, False, False, False, False, False])
        constraint.is_valid = Mock(return_value=is_valid_result)

        # Run / Assert
        error_message = re.escape(
            "Data is not valid for the 'Constraint' constraint:\n   "
            'a  b\n1  1  4\n3  3  6\n4  4  7\n5  5  8\n6  6  9'
            '\n+1 more'
        )
        with pytest.raises(ConstraintsNotMetError, match=error_message):
            constraint._validate_data_meets_constraint(data)

    def test__validate_data_meets_constraints_missing_cols(self):
        """Test the ``_validate_data_meets_constraint`` method.

        Expect that the method doesn't do anything when the columns are not in the given data.

        Input:
        - Table data that is missing a constraint column
        Output:
        - None
        Side Effects:
        - No error
        """
        # Setup
        data = pd.DataFrame({
            'a': [0, 1, 2],
            'b': [3, 4, 5]
        }, index=[0, 1, 2])
        constraint = Constraint()
        constraint.constraint_columns = ['a', 'b', 'c']
        constraint.is_valid = Mock()

        # Run
        constraint._validate_data_meets_constraint(data)

        # Assert
        assert not constraint.is_valid.called

    def test_transform(self):
        """Test the ``Constraint.transform`` method.

        By default, it behaves like an identity method, to be optionally overwritten by subclasses.

        The ``Constraint.transform`` method is expected to:
            - Return a copy of the input data.

        Input:
            - a DataFrame

        Output:
            - Input
        """
        # Setup
        instance = Constraint()
        data = pd.DataFrame({'col': ['input']})

        # Run
        output = instance.transform(data)

        # Assert
        pd.testing.assert_frame_equal(output, pd.DataFrame({'col': ['input']}))
        assert id(output) != id(data)

    def test_transform_calls__transform(self):
        """Test that the ``Constraint.transform`` method calls ``_transform``.

        The ``Constraint.transform`` method is expected to:
            - Return value returned by ``_transform``.

        Input:
            - Anything

        Output:
            - Result of ``_transform(input)``
        """
        # Setup
        constraint_mock = Mock()
        constraint_mock.constraint_columns = []
        constraint_mock._transform.return_value = 'the_transformed_data'

        # Run
        output = Constraint.transform(constraint_mock, pd.DataFrame())

        # Assert
        assert output == 'the_transformed_data'

    def test_transform__transform_errors(self):
        """Test that the ``transform`` method handles any errors.

        If the ``_transform`` method raises an error, the error should be raised.

        Setup:
            - Make ``_transform`` raise an error.

        Input:
            - ``pandas.DataFrame``.

        Output:
            - Same ``pandas.DataFrame``.

        Side effects:
            - Exception should be raised
        """
        # Setup
        instance = Constraint()
        instance._transform = Mock()
        instance._transform.side_effect = Exception()
        data = pd.DataFrame({'a': [1, 2, 3]})

        # Run / Assert
        with pytest.raises(Exception):
            instance.transform(data)

    def test_transform_columns_missing(self):
        """Test the ``Constraint.transform`` method with invalid data.

        If ``table_data`` is missing any columns it should raise a
        ``MissingConstraintColumnError``.

        The ``Constraint.transform`` method is expected to:
        - Raise ``MissingConstraintColumnError``.
        """
        # Run
        instance = Constraint()
        instance._transform = lambda x: x
        instance.constraint_columns = ('a',)

        # Assert
        with pytest.raises(MissingConstraintColumnError):
            instance.transform(pd.DataFrame([[1, 2], [3, 4]], columns=['b', 'c']))

    def test_transform_all_columns_missing(self):
        """Test the ``Constraint.transform`` method with all columns missing.

        If ``table_data`` is missing all of the ``constraint_columns`` a
        ``MissingConstraintColumnError`` is raised.

        The ``Constraint.transform`` method is expected to:
        - Raise ``MissingConstraintColumnError``.
        """
        # Run
        instance = Constraint()
        instance._transform = lambda x: x
        instance.constraint_columns = ('a',)

        # Assert
        with pytest.raises(MissingConstraintColumnError):
            instance.transform(pd.DataFrame())

    def test_fit_transform(self):
        """Test the ``Constraint.fit_transform`` method.

        The ``Constraint.fit_transform`` method is expected to:
        - Call the ``fit`` method.
        - Call the ``transform`` method.
        - Return the input data unmodified.

        Input:
        - Anything
        Output:
        - self.transform output
        Side Effects:
        - self.fit is called with input
        - self.transform is called with input
        """
        # Setup
        constraint_mock = Mock()
        constraint_mock.transform.return_value = 'the_transformed_data'

        # Run
        data = 'my_data'
        output = Constraint.fit_transform(constraint_mock, data)

        # Assert
        assert output == 'the_transformed_data'

        constraint_mock.fit.assert_called_once_with('my_data')
        constraint_mock.transform.assert_called_once_with('my_data')

    def test_reverse_transform(self):
        """Test the ``Constraint.reverse_transform`` method. It is an identity method
        for completion, to be optionally overwritten by subclasses.

        The ``Constraint.reverse_transform`` method is expected to:
            - Return a copy of the input data.

        Input:
            - Anything
        Output:
            - Input
        """
        # Setup
        instance = Constraint()
        data = pd.DataFrame()

        # Run
        output = instance.reverse_transform(data)

        # Assert
        pd.testing.assert_frame_equal(output, pd.DataFrame())
        assert id(output) != id(data)

    def test_is_valid(self):
        """Test the ``Constraint.is_valid` method. This should be overwritten by all the
        subclasses that have a way to decide which rows are valid and which are not.

        The ``Constraint.is_valid`` method is expected to:
        - Say whether the given table rows are valid.

        Input:
        - Table data (pandas.DataFrame)
        Output:
        - Series of ``True`` values (pandas.Series)
        """
        # Setup
        table_data = pd.DataFrame({
            'a': [1, 2, 3]
        })

        # Run
        instance = Constraint()
        out = instance.is_valid(table_data)

        # Assert
        expected_out = pd.Series([True, True, True])
        pd.testing.assert_series_equal(expected_out, out)

    def test_filter_valid(self):
        """Test the ``Constraint.filter_valid`` method.

        The ``Constraint.filter_valid`` method is expected to:
        - Filter the input data by calling the method ``is_valid``.

        Input:
        - Table data (pandas.DataFrame)
        Output:
        - Table data, with only the valid rows (pandas.DataFrame)
        """
        # Setup
        table_data = pd.DataFrame({
            'a': [1, 2, 3]
        })

        constraint_mock = Mock()
        constraint_mock.is_valid.return_value = pd.Series([True, True, False])

        # Run
        out = Constraint.filter_valid(constraint_mock, table_data)

        # Assert
        expected_out = pd.DataFrame({
            'a': [1, 2]
        })
        pd.testing.assert_frame_equal(expected_out, out)

    def test_filter_valid_with_invalid_index(self):
        """Test the ``Constraint.filter_valid`` method.

        Tests when the is_valid method returns a Series with an invalid index.

        Note: `is_valid.index` can be [0, 1, 5] if, for example, the Series is a subset
        of an original table with 10 rows, but only rows 0/1/5 were selected.

        Input:
        - Table data (pandas.DataFrame)
        Output:
        - Table data, with only the valid rows (pandas.DataFrame)
        """
        # Setup
        table_data = pd.DataFrame({
            'a': [1, 2, 3]
        })

        constraint_mock = Mock()
        is_valid = pd.Series([True, True, False])
        is_valid.index = [0, 1, 5]
        constraint_mock.is_valid.return_value = is_valid

        # Run
        out = Constraint.filter_valid(constraint_mock, table_data)

        # Assert
        expected_out = pd.DataFrame({
            'a': [1, 2]
        })
        pd.testing.assert_frame_equal(expected_out, out)

    def test_from_dict_fqn(self):
        """Test the ``Constraint.from_dict`` method passing a FQN.

        If the ``constraint`` string is a FQN, import the class
        before creating an instance of it.

        Input:
        - constraint dict with a FQN and args
        Output:
        - Instance of the subclass with the right args.
        """
        # Setup
        constraint_dict = {
            'constraint': 'sdv.constraints.tabular.FixedCombinations',
            'column_names': ['a', 'b'],
        }

        # Run
        instance = Constraint.from_dict(constraint_dict)

        # Assert
        assert isinstance(instance, FixedCombinations)
        assert instance._columns == ['a', 'b']

    def test_from_dict_subclass(self):
        """Test the ``Constraint.from_dict`` method passing a subclass name.

        If the ``constraint`` string is a subclass name, take it from the
        Subclasses dict.

        Input:
        - constraint dict with a subclass name and args
        Output:
        - Instance of the subclass with the right args.
        """
        # Setup
        constraint_dict = {
            'constraint': 'FixedCombinations',
            'column_names': ['a', 'b'],
        }

        # Run
        instance = Constraint.from_dict(constraint_dict)

        # Assert
        assert isinstance(instance, FixedCombinations)
        assert instance._columns == ['a', 'b']

    def test_to_dict(self):
        """Test the ``Constraint.to_dict`` method.

        The ``Constraint.to_dict`` method is expected to return a dict
        containting the FQN of the constraint instance and all the
        required arguments rebuild it.

        Output:
        - Dict with the right values.
        """
        # Run
        instance = FixedCombinations(column_names=['a', 'b'])
        constraint_dict = instance.to_dict()

        # Assert
        expected_dict = {
            'constraint': 'sdv.constraints.tabular.FixedCombinations',
            'column_names': ['a', 'b'],
        }
        assert constraint_dict == expected_dict


class TestColumnsModel:

    def test___init__(self):
        """Test the ``__init__`` method of ``ColumnsModel``.

        Test that creating a ``ColumnsModel`` with a column name, stores it as a list.

        Setup:
            - Create an instance of ``ColumnsModel`` with a string representing the column
              name.

        Input:
            - String name of a column.

        Side Effects:
            - ``instance.constraint_columns`` is a list from the string given before.
        """

        # Run
        constraint = Mock()
        instance = ColumnsModel(constraint, 'age')

        # Assert
        assert instance.constraint_columns == ['age']
        assert instance.constraint == constraint

    def test___init__list(self):
        """Test the ``__init__`` method of ``ColumnsModel``.

        Test that creating a ``ColumnsModel`` with a list, stores it as a list

        Setup:
            - Create an instance of ``ColumnsModel`` with a list of strings.

        Input:
            - List of columns.

        Side Effects:
            - ``instance.constraint_columns`` is the input list.
        """

        # Run
        constraint = Mock()
        instance = ColumnsModel(constraint, ['age', 'age_when_joined'])

        # Assert
        assert instance.constraint_columns == ['age', 'age_when_joined']
        assert instance.constraint == constraint

    @patch('sdv.constraints.base.FloatFormatter')
    @patch('sdv.constraints.base.OneHotEncoder')
    @patch('sdv.constraints.base.UnixTimestampEncoder')
    @patch('sdv.constraints.base.BinaryEncoder')
    def test__get_hyper_transformer_config(self, mock_binaryencoder, mock_unixtimestampencoder,
                                           mock_onehotencoder, mock_floatformatter):
        """Test the ``_get_hyper_transformer_config``.

        Test that the method ``_get_hyper_transformer_config`` returns the expected
        ``sdtypes`` and ``transformers`` for a given ``data_to_model``.

        Setup:
            - Create a ``pandas.DataFrame`` named ``data_to_model`` whit multiple ``dtypes``.

        Mock:
            - ``rdt.transformers`` from ``constraints.base``.

        Input:
            - ``data_to_model`` with boolean, datetime, numerical and categorical data.

        Output:
            - A dictionary containing ``sdtypes`` and ``transformers`` that match the expected
              to the ``sdtypes``.
        """
        # Setup
        data_to_model = pd.DataFrame({
            'age': [1, 2, 3],
            'amount': [1, 2, None],
            'name': [None, 'Doe', 'John Doe'],
            'joindate': pd.to_datetime(['2021-02-05', None, '2021-12-21']),
            'is_valid': [True, False, None],
        })
        age_float_formatter = Mock()
        amount_float_formatter = Mock()
        mock_floatformatter.side_effects = [age_float_formatter, amount_float_formatter]

        # Run
        ht_config = ColumnsModel._get_hyper_transformer_config(data_to_model)

        # Assert
        ht_config == {
            'sdtypes': {
                'age': 'numerical',
                'amount': 'numerical',
                'name': 'categorical',
                'joindate': 'datetime',
                'is_valid': 'boolean'
            },
            'transformers': {
                'age': age_float_formatter,
                'amount': amount_float_formatter,
                'name': mock_onehotencoder,
                'joindate': mock_unixtimestampencoder.return_value,
                'is_valid': mock_binaryencoder.return_value
            }

        }

    @patch('sdv.constraints.base.GaussianMultivariate')
    @patch('sdv.constraints.base.HyperTransformer')
    def test_fit(self, mock_hyper_transformer, mock_gaussian_multivariate):
        """Test the ``fit`` method.

        The ``fit`` method should create an instance of ``rdt.HyperTransformer`` and use
        the ``fit_transform`` method in order to transform the data, which afterwards is being
        fitted to the ``GaussianMultivariate`` which uses as ``distribution`` the
        ``GaussianUnivariate``.

        Setup:
            - Instance of ``ColumnsModel``.

        Input:
            - table_data with 3 columns.

        Mock:
            - Mock ``rdt.HyperTransformer``.
            - Mock ``GaussianMultivariate``.

        Side Effects:
            - Instance of ``rdt.HyperTransformer`` has been created and fitted.
            - A ``GaussianMultivariate`` model has been created and fitted.
        """
        # Setup
        table_data = pd.DataFrame({
            'age': [1, 2, 3, 4],
            'age_when_joined': [5, 6, 7, 8],
            'retirement': ['a', 'b', 'c', 'd']
        })

        mock_hyper_transformer.return_value.fit_transform.return_value = 'transformed_data'
        instance = ColumnsModel(Mock(), ['age', 'age_when_joined'])

        # Run
        instance.fit(table_data)

        # Assert
        mock_hyper_transformer.assert_called_once_with()
        mock_hyper_transformer.return_value.set_config.assert_called_once()
        call_data = mock_hyper_transformer.return_value.fit_transform.call_args[0][0]
        pd.testing.assert_frame_equal(table_data[['age', 'age_when_joined']], call_data)

        mock_gaussian_multivariate.assert_called_once_with(distribution=GaussianUnivariate)
        mock_gaussian_multivariate.return_value.fit.assert_called_once_with('transformed_data')
        assert instance._hyper_transformer
        assert instance._model

    def test__reject_sample(self):
        """Test the ``ColumnsModel._reject_sample``.

        The ``_reject_sample`` method should call the ``instance._model`` wiith the given
        number of rows and ``conditions``.

        Setup:
            - Instance of ``ColumnsModel.

        Mock:
            - ``instance._hyper_transformer``.
            - ``instance._model``.

        Input:
            - Table with some missing columns.

        Output:
            - Transformed data with all columns.
        """
        # Setup
        constraint = Mock()
        constraint.is_valid.side_effect = lambda x: pd.Series(
            [True for _ in range(len(x))],
            index=x.index
        )
        instance = ColumnsModel(constraint, ['a', 'b'])
        instance._hyper_transformer = Mock()
        instance._model = Mock()
        transformed_conditions = [pd.DataFrame([[1], [1], [1], [1], [1]], columns=['b'])]
        instance._model.sample.side_effect = [
            pd.DataFrame([
                [1, 2],
                [1, 3]
            ], columns=['a', 'b']),
            pd.DataFrame([
                [1, 4],
                [1, 5],
                [1, 6],
                [1, 7]
            ], columns=['a', 'b']),
        ]
        instance._hyper_transformer.transform.side_effect = transformed_conditions
        instance._hyper_transformer.reverse_transform = lambda x: x

        # Run
        transformed_data = instance._reject_sample(num_rows=5, conditions={'b': 1})

        # Assert
        expected_result = pd.DataFrame([
            [1, 2],
            [1, 3],
            [1, 4],
            [1, 5],
            [1, 6]
        ], columns=['a', 'b'])
        model_calls = instance._model.sample.mock_calls
        assert len(model_calls) == 2
        instance._model.sample.assert_any_call(num_rows=5, conditions={'b': 1})
        assert model_calls[1][2]['num_rows'] > 3
        pd.testing.assert_frame_equal(transformed_data, expected_result)

        expected_call_1 = pd.DataFrame({'a': [1, 1], 'b': [2, 3]})
        expected_call_2 = pd.DataFrame([
            [1, 4],
            [1, 5],
            [1, 6],
            [1, 7]
        ], columns=['a', 'b'])

        pd.testing.assert_frame_equal(expected_call_1, constraint.is_valid.call_args_list[0][0][0])
        pd.testing.assert_frame_equal(expected_call_2, constraint.is_valid.call_args_list[1][0][0])

    def test__reject_sampling_duplicates_valid_rows(self):
        """Test the ``Constraint.transform`` method's reject sampling fall back.

        After 100 tries, some valid rows are created but not enough, then the valid rows
        are duplicated to meet the ``num_rows`` requirement.

        Setup:
            - Instance of ``ColumnsModel.

        Mock:
            - ``instance._hyper_transformer``.
            - ``instance._model``.

        Input:
            - Table with some missing columns.

        Output:
            - Transformed data with all columns.
        """
        # Setup
        constraint = Mock()
        constraint.is_valid.side_effect = lambda x: pd.Series(
            [True for _ in range(len(x))],
            index=x.index
        )
        instance = ColumnsModel(constraint, ['a', 'b'])
        instance._hyper_transformer = Mock()
        instance._model = Mock()
        transformed_conditions = [pd.DataFrame([[1], [1], [1], [1], [1]], columns=['b'])]
        instance._model.sample.side_effect = [
            pd.DataFrame([
                [1, 2],
                [1, 3]
            ], columns=['a', 'b'])
        ] + [pd.DataFrame()] * 100
        instance._hyper_transformer.transform.side_effect = transformed_conditions
        instance._hyper_transformer.reverse_transform = lambda x: x

        # Run
        transformed_data = instance._reject_sample(num_rows=5, conditions={'b': 1})

        # Assert
        expected_result = pd.DataFrame([
            [1, 2],
            [1, 3],
            [1, 2],
            [1, 3],
            [1, 2]
        ], columns=['a', 'b'])
        model_calls = instance._model.sample.mock_calls
        assert len(model_calls) == 101
        instance._model.sample.assert_any_call(num_rows=5, conditions={'b': 1})
        pd.testing.assert_frame_equal(transformed_data, expected_result)

    def test__reject_sampling_no_valid_rows(self):
        """Test the ``ColumnsModel._reject_sampling`` method.

        Raises an error when there is no valid rows within 100 tries.

        Setup:
            - Instance of ``ColumnsModel.

        Mock:
            - ``instance._hyper_transformer``.
            - ``instance._model``.
            - ``instance.constraint``.

        Input:
            - Table with some missing columns.

        Side Effects:
            - A ``ValueError`` should be raised.
        """
        # Setup
        constraint = Mock()
        constraint.is_valid.side_effect = lambda x: pd.Series(
            [False for _ in range(len(x))],
            index=x.index
        )
        instance = ColumnsModel(constraint, ['a', 'b'])
        instance._hyper_transformer = Mock()
        instance._model = Mock()
        transformed_conditions = [pd.DataFrame([[1], [1], [1], [1], [1]], columns=['b'])]
        instance._model.sample.side_effect = [
            pd.DataFrame([
                [1, 2],
                [1, 3]
            ], columns=['a', 'b'])
        ] + [pd.DataFrame()] * 100
        instance._hyper_transformer.transform.side_effect = transformed_conditions
        instance._hyper_transformer.reverse_transform = lambda x: x

        # Run
        with pytest.raises(ValueError, match='Could not get enough valid rows within 100 trials.'):
            instance._reject_sample(num_rows=5, conditions={'b': 1})

    def test_sample(self):
        """Test the ``ColumnsModel.sample`` method.

        This method should call ``instance._reject_sample`` with the given
        conditions.

        Setup:
            - Instance of ``ColumnsModel.

        Mock:
            - ``instance._hyper_transformer``.
            - ``instance._reject_sample``.

        Input:
            - Table with some missing columns.

        Output:
            - Transformed data with all columns.
        """
        # Setup
        constraint = Mock()
        instance = ColumnsModel(constraint, ['a', 'b'])
        instance._hyper_transformer = Mock()
        instance._model = Mock()
        transformed_conditions = [pd.DataFrame([[1], [1], [1], [1], [1]], columns=['b'])]
        instance._hyper_transformer.transform.side_effect = transformed_conditions
        instance._hyper_transformer.reverse_transform = lambda x: x
        instance._reject_sample = Mock()
        instance._reject_sample.side_effect = [
            pd.DataFrame([
                [1, 2],
                [1, 3],
                [1, 4],
                [1, 5],
                [1, 6],
            ], columns=['a', 'b']),
        ]

        # Run
        data = pd.DataFrame([[1], [1], [1], [1], [1]], columns=['b'])
        transformed_data = instance.sample(data)

        # Assert
        expected_result = pd.DataFrame([
            [1, 2],
            [1, 3],
            [1, 4],
            [1, 5],
            [1, 6]
        ], columns=['a', 'b'])
        instance._reject_sample.assert_any_call(num_rows=5, conditions={'b': 1})
        pd.testing.assert_frame_equal(transformed_data, expected_result)



"""Combination of GaussianCopula transformation and GANs."""

from rdt import HyperTransformer
from rdt.transformers import GaussianNormalizer

from sdv.tabular.ctgan import CTGAN


class CopulaGAN(CTGAN):
    """Combination of GaussianCopula transformation and GANs.

    This model extends the ``CTGAN`` model to add the flexibility of the GaussianCopula
    transformations provided by the ``GaussianNormalizer`` from ``RDT``.

    Overall, the fitting process consists of the following steps:

    1. Transform each non categorical variable from the input
       data using a ``GaussianNormalizer``:

       i. If not specified, find out the distribution which each one
          of the variables from the input dataset has.
       ii. Transform each variable to a standard normal space by applying
           the CDF of the corresponding distribution and later on applying
           an inverse CDF from a standard normal distribution.

    2. Fit CTGAN with the transformed table.

    And the process of sampling is:

    1. Sample using CTGAN
    2. Reverse the previous transformation by applying the CDF of a standard normal
       distribution and then inverting the CDF of the distribution that correpsonds
       to each variable.

    The arguments of this model are the same as for CTGAN except for two additional
    arguments, ``field_distributions`` and ``default_distribution`` that give the
    ability to define specific transformations for individual fields as well as
    which distribution to use by default if no specific distribution has been selected.

    Distributions can be passed as a ``copulas`` univariate instance or as one
    of the following string values:

    * ``gaussian``: Use a Gaussian distribution.
    * ``gamma``: Use a Gamma distribution.
    * ``beta``: Use a Beta distribution.
    * ``student_t``: Use a Student T distribution.
    * ``gaussian_kde``: Use a GaussianKDE distribution. This model is non-parametric,
      so using this will make ``get_parameters`` unusable.
    * ``truncated_gaussian``: Use a Truncated Gaussian distribution.

    Args:
        field_names (list[str]):
            List of names of the fields that need to be modeled
            and included in the generated output data. Any additional
            fields found in the data will be ignored and will not be
            included in the generated output.
            If ``None``, all the fields found in the data are used.
        field_types (dict[str, dict]):
            Dictinary specifying the data types and subtypes
            of the fields that will be modeled. Field types and subtypes
            combinations must be compatible with the SDV Metadata Schema.
        field_transformers (dict[str, str]):
            Dictinary specifying which transformers to use for each field.
            Available transformers are:

                * ``FloatFormatter``: Uses a ``FloatFormatter`` for numerical data.
                * ``FrequencyEncoder``: Uses a ``FrequencyEncoder`` without gaussian noise.
                * ``FrequencyEncoder_noised``: Uses a ``FrequencyEncoder`` adding gaussian noise.
                * ``OneHotEncoder``: Uses a ``OneHotEncoder``.
                * ``LabelEncoder``: Uses a ``LabelEncoder`` without gaussian nose.
                * ``LabelEncoder_noised``: Uses a ``LabelEncoder`` adding gaussian noise.
                * ``BinaryEncoder``: Uses a ``BinaryEncoder``.
                * ``UnixTimestampEncoder``: Uses a ``UnixTimestampEncoder``.

        anonymize_fields (dict[str, str]):
            Dict specifying which fields to anonymize and what faker
            category they belong to.
        primary_key (str):
            Name of the field which is the primary key of the table.
        constraints (list[Constraint, dict]):
            List of Constraint objects or dicts.
        table_metadata (dict or metadata.Table):
            Table metadata instance or dict representation.
            If given alongside any other metadata-related arguments, an
            exception will be raised.
            If not given at all, it will be built using the other
            arguments or learned from the data.
        log_frequency (boolean):
            Whether to use log frequency of categorical levels in conditional
            sampling. Defaults to ``True``.
        embedding_dim (int):
            Size of the random sample passed to the Generator. Defaults to 128.
        generator_dim (tuple or list of ints):
            Size of the output samples for each one of the Residuals. A Resiudal Layer
            will be created for each one of the values provided. Defaults to (256, 256).
        discriminator_dim (tuple or list of ints):
            Size of the output samples for each one of the Discriminator Layers. A Linear
            Layer will be created for each one of the values provided. Defaults to (256, 256).
        batch_size (int):
            Number of data samples to process in each step.
        verbose (bool):
            Whether to print fit progress on stdout. Defaults to ``False``.
        epochs (int):
            Number of training epochs. Defaults to 300.
        cuda (bool or str):
            If ``True``, use CUDA. If an ``str``, use the indicated device.
            If ``False``, do not use cuda at all.
        field_distributions (dict):
            Optionally specify a dictionary that maps the name of each field to the distribution
            that must be used in it. Fields that are not specified in the input ``dict`` will
            be modeled using the default distribution. Defaults to ``None``.
        default_distribution (copulas.univariate.Univariate or str):
            Distribution to use on the fields for which no specific distribution has been given.
            Defaults to ``truncated_gaussian``.
        learn_rounding_scheme (bool):
            Define rounding scheme for ``FloatFormatter``. If ``True``, the data returned by
            ``reverse_transform`` will be rounded to that place. Defaults to ``True``.
        enforce_min_max_values (bool):
            Specify whether or not to clip the data returned by ``reverse_transform`` of
            the numerical transformer, ``FloatFormatter``, to the min and max values seen
            during ``fit``. Defaults to ``True``.
    """

    DEFAULT_DISTRIBUTION = 'truncated_gaussian'
    _field_distributions = None
    _default_distribution = None
    _hyper_transformer = None

    def __init__(self, field_names=None, field_types=None, field_transformers=None,
                 anonymize_fields=None, primary_key=None, constraints=None, table_metadata=None,
                 embedding_dim=128, generator_dim=(256, 256), discriminator_dim=(256, 256),
                 generator_lr=2e-4, generator_decay=1e-6, discriminator_lr=2e-4,
                 discriminator_decay=1e-6, batch_size=500, discriminator_steps=1,
                 log_frequency=True, verbose=False, epochs=300, cuda=True,
                 field_distributions=None, default_distribution=None, learn_rounding_scheme=True,
                 enforce_min_max_values=True):
        super().__init__(
            field_names=field_names,
            primary_key=primary_key,
            field_types=field_types,
            field_transformers=field_transformers,
            anonymize_fields=anonymize_fields,
            constraints=constraints,
            table_metadata=table_metadata,
            embedding_dim=embedding_dim,
            generator_dim=generator_dim,
            discriminator_dim=discriminator_dim,
            generator_lr=generator_lr,
            generator_decay=generator_decay,
            discriminator_lr=discriminator_lr,
            discriminator_decay=discriminator_decay,
            batch_size=batch_size,
            discriminator_steps=discriminator_steps,
            log_frequency=log_frequency,
            verbose=verbose,
            epochs=epochs,
            cuda=cuda,
            learn_rounding_scheme=learn_rounding_scheme,
            enforce_min_max_values=enforce_min_max_values,
        )
        self._field_distributions = field_distributions or dict()
        self._default_distribution = default_distribution or self.DEFAULT_DISTRIBUTION

    def get_distributions(self):
        """Get the marginal distributions used by this CopulaGAN.

        Returns:
            dict:
                Dictionary containing the distributions used or detected
                for each column.
        """
        return {
            transformer.column_prefix: transformer._univariate.to_dict()['type']
            for transformer in self._hyper_transformer._transformers_sequence
            if isinstance(transformer, GaussianNormalizer)
        }

    def _fit(self, table_data):
        """Fit the model to the table.

        Args:
            table_data (pandas.DataFrame):
                Data to be learned.
        """
        distributions = self._field_distributions
        fields = self._metadata.get_fields()

        sdtypes = {}
        transformers = {}
        for field in table_data:
            field_name = field.replace('.value', '')
            field_sdtype = fields.get(field_name, {}).get('type')
            if field_name in fields and field_sdtype != 'categorical':
                sdtypes[field] = 'numerical'
                transformers[field] = GaussianNormalizer(
                    model_missing_values=True,
                    distribution=distributions.get(field_name, self._default_distribution)
                )
            else:
                sdtypes[field] = field_sdtype or 'categorical'
                transformers[field] = None

        self._hyper_transformer = HyperTransformer()
        self._hyper_transformer.set_config({'transformers': transformers, 'sdtypes': sdtypes})
        table_data = self._hyper_transformer.fit_transform(table_data[list(transformers)])

        super()._fit(table_data)

    def _sample(self, num_rows, conditions=None):
        """Sample the indicated number of rows from the model.

        Args:
            num_rows (int):
                Amount of rows to sample.
            conditions (dict):
                If specified, this dictionary maps column names to the column
                value. Then, this method generates `num_rows` samples, all of
                which are conditioned on the given variables.

        Returns:
            pandas.DataFrame:
                Sampled data.
        """
        sampled = super()._sample(num_rows, conditions)
        return self._hyper_transformer.reverse_transform(sampled)



"""Timeseries models based on the DeepEcho library.

This modulde defines a base DeepEchoModel class which prepares the
data in the format that the DeepEcho models expect.
This class is also responsible for transforming the sequence index
in a way that can be learned by the models and for transforming it
back to the original data format after sampling.

Currently implemented models are:
    - PAR: Based on the deepecho.models.par.PARModel
"""

import numpy as np
import pandas as pd
import tqdm
from deepecho import PARModel
from deepecho.sequences import assemble_sequences

from sdv.timeseries.base import BaseTimeseriesModel


class DeepEchoModel(BaseTimeseriesModel):
    """Base class for all the SDV Time series models based on DeepEcho."""

    _MODEL_CLASS = None
    _model_kwargs = None
    _verbose = False

    def _build_model(self):
        return self._MODEL_CLASS(**self._model_kwargs)

    def _transform_sequence_index(self, sequences):
        sequence_index_idx = self._data_columns.index(self._sequence_index)
        for sequence in sequences:
            data = sequence['data']
            sequence_index = data[sequence_index_idx]
            diffs = np.diff(sequence_index).tolist()
            data[sequence_index_idx] = diffs[0:1] + diffs
            data.append(sequence_index[0:1] * len(sequence_index))

    def _fit(self, timeseries_data):
        self._model = self._build_model()

        if self._sequence_index:
            timeseries_data = timeseries_data.rename(columns={
                self._sequence_index + '.value': self._sequence_index
            })

        self._output_columns = list(timeseries_data.columns)
        self._data_columns = [
            column
            for column in timeseries_data.columns
            if column not in self._entity_columns + self._context_columns
        ]

        sequences = assemble_sequences(
            timeseries_data,
            self._entity_columns,
            self._context_columns,
            self._segment_size,
            self._sequence_index,
            drop_sequence_index=False
        )

        data_types = list()
        context_types = list()
        for field in self._output_columns:
            dtype = timeseries_data[field].dtype
            kind = dtype.kind
            if kind in ('i', 'f'):
                data_type = 'continuous'
            elif kind in ('O', 'b'):
                data_type = 'categorical'
            else:
                raise ValueError(f'Unsupported dtype {dtype}')

            if field in self._data_columns:
                data_types.append(data_type)
            elif field in self._context_columns:
                context_types.append(data_type)

        if self._sequence_index:
            self._transform_sequence_index(sequences)
            data_types.append('continuous')

        # Validate and fit
        self._model.fit_sequences(sequences, context_types, data_types)

    def _sample(self, context=None, sequence_length=None):
        """Sample new sequences.

        Args:
            context (pandas.DataFrame):
                Context values to use when generating the sequences.
                If not passed, the context values will be sampled
                using the specified tabular model.
            sequence_length (int):
                If passed, sample sequences of this length. If not
                given, the sequence length will be sampled from
                the model.

        Returns:
            pandas.DataFrame:
                Table containing the sampled sequences in the same
                format as that he training data had.
        """
        # Set the entity_columns as index to properly iterate over them
        if self._entity_columns:
            context = context.set_index(self._entity_columns)

        iterator = tqdm.tqdm(context.iterrows(), disable=not self._verbose, total=len(context))

        output = list()
        for entity_values, context_values in iterator:
            context_values = context_values.tolist()
            sequence = self._model.sample_sequence(context_values, sequence_length)
            if self._sequence_index:
                sequence_index_idx = self._data_columns.index(self._sequence_index)
                diffs = sequence[sequence_index_idx]
                start = sequence.pop(-1)
                sequence[sequence_index_idx] = np.cumsum(diffs) - diffs[0] + start

            # Reformat as a DataFrame
            group = pd.DataFrame(
                dict(zip(self._data_columns, sequence)),
                columns=self._data_columns
            )
            group[self._entity_columns] = entity_values
            for column, value in zip(self._context_columns, context_values):
                if column == self._sequence_index:
                    sequence_index = group[column]
                    group[column] = sequence_index.cumsum() - sequence_index.iloc[0] + value
                else:
                    group[column] = value

            output.append(group)

        output = pd.concat(output)
        output = output[self._output_columns].reset_index(drop=True)
        if self._sequence_index:
            output = output.rename(columns={
                self._sequence_index: self._sequence_index + '.value'
            })

        return output


class PAR(DeepEchoModel):
    """DeepEcho model based on the deepecho.models.par.PARModel class.

    Args:
        field_names (list[str]):
            List of names of the fields that need to be modeled
            and included in the generated output data. Any additional
            fields found in the data will be ignored and will not be
            included in the generated output.
            If ``None``, all the fields found in the data are used.
        field_types (dict[str, dict]):
            Dictinary specifying the data types and subtypes
            of the fields that will be modeled. Field types and subtypes
            combinations must be compatible with the SDV Metadata Schema.
        anonymize_fields (dict[str, str]):
            Dict specifying which fields to anonymize and what faker
            category they belong to.
        primary_key (str):
            Name of the field which is the primary key of the table.
        entity_columns (list[str]):
            Names of the columns which identify different time series
            sequences. These will be used to group the data in separated
            training examples.
        context_columns (list[str]):
            The columns in the dataframe which are constant within each
            group/entity. These columns will be provided at sampling time
            (i.e. the samples will be conditioned on the context variables).
        segment_size (int, pd.Timedelta or str):
            If specified, cut each training sequence in several segments of
            the indicated size. The size can either can passed as an integer
            value, which will interpreted as the number of data points to
            put on each segment, or as a pd.Timedelta (or equivalent str
            representation), which will be interpreted as the segment length
            in time. Timedelta segment sizes can only be used with sequence
            indexes of type datetime.
        sequence_index (str):
            Name of the column that acts as the order index of each
            sequence. The sequence index column can be of any type that can
            be sorted, such as integer values or datetimes.
        context_model (str or sdv.tabular.BaseTabularModel):
            Model to use to sample the context rows. It can be passed as a
            a string, which must be one of the following:

            * `gaussian_copula` (default): Use a GaussianCopula model.

            Alternatively, a preconfigured Tabular model instance can be
            passed.

        table_metadata (dict or metadata.Table):
            Table metadata instance or dict representation.
            If given alongside any other metadata-related arguments, an
            exception will be raised.
            If not given at all, it will be built using the other
            arguments or learned from the data.
        epochs (int):
            The number of epochs to train for. Defaults to 128.
        sample_size (int):
            The number of times to sample (before choosing and
            returning the sample which maximizes the likelihood).
            Defaults to 1.
        cuda (bool):
            Whether to attempt to use cuda for GPU computation.
            If this is False or CUDA is not available, CPU will be used.
            Defaults to ``True``.
        verbose (bool):
            Whether to print progress to console or not.

    """

    _MODEL_CLASS = PARModel

    def __init__(self, field_names=None, field_types=None, anonymize_fields=None,
                 primary_key=None, entity_columns=None, context_columns=None,
                 sequence_index=None, segment_size=None, context_model=None,
                 table_metadata=None, epochs=128, sample_size=1, cuda=True, verbose=False):
        super().__init__(
            field_names=field_names,
            field_types=field_types,
            anonymize_fields=anonymize_fields,
            primary_key=primary_key,
            entity_columns=entity_columns,
            context_columns=context_columns,
            sequence_index=sequence_index,
            segment_size=segment_size,
            context_model=context_model,
            table_metadata=table_metadata,
        )

        self._model_kwargs = {
            'epochs': epochs,
            'sample_size': sample_size,
            'cuda': cuda,
            'verbose': verbose,
        }
        self._verbose = verbose



from unittest.mock import MagicMock, Mock, call, patch

import graphviz
import pytest

from sdv.metadata import Metadata, visualization


def test__get_graphviz_extension_path_without_extension():
    """Raises a ValueError when the path doesn't contains an extension."""
    with pytest.raises(ValueError):
        visualization._get_graphviz_extension('/some/path')


def test__get_graphviz_extension_invalid_extension():
    """Raises a ValueError when the path contains an invalid extension."""
    with pytest.raises(ValueError):
        visualization._get_graphviz_extension('/some/path.foo')


def test__get_graphviz_extension_none():
    """Get graphviz with path equals to None."""
    # Run
    result = visualization._get_graphviz_extension(None)

    # Asserts
    assert result == (None, None)


def test__get_graphviz_extension_valid():
    """Get a valid graphviz extension."""
    # Run
    result = visualization._get_graphviz_extension('/some/path.png')

    # Asserts
    assert result == ('/some/path', 'png')


def test__add_nodes():
    """Add nodes into a graphviz digraph."""
    # Setup
    metadata = MagicMock(spec_set=Metadata)
    minimock = Mock()

    # pass tests in python3.5
    minimock.items.return_value = (
        ('a_field', {'type': 'numerical', 'subtype': 'integer'}),
        ('b_field', {'type': 'id'}),
        ('c_field', {'type': 'id', 'ref': {'table': 'other', 'field': 'pk_field'}})
    )

    metadata.get_tables.return_value = ['demo']
    metadata.get_fields.return_value = minimock

    metadata.get_primary_key.return_value = 'b_field'
    metadata.get_parents.return_value = set(['other'])
    metadata.get_foreign_keys.return_value = ['c_field']

    metadata.get_table_meta.return_value = {'path': None}

    plot = Mock()

    # Run
    visualization._add_nodes(metadata, plot, names=True, details=True)

    # Asserts
    expected_node_label = (
        r"{demo|a_field : numerical - integer\lb_field : id\l"
        r"c_field : id\l|Primary key: b_field\l"
        r"Foreign key (other): c_field\l}"
    )

    metadata.get_fields.assert_called_once_with('demo')
    metadata.get_primary_key.assert_called_once_with('demo')
    metadata.get_parents.assert_called_once_with('demo')
    metadata.get_table_meta.assert_called_once_with('demo')
    metadata.get_foreign_keys.assert_called_once_with('other', 'demo')
    metadata.get_table_meta.assert_called_once_with('demo')

    plot.node.assert_called_once_with('demo', label=expected_node_label)


def test__add_edges():
    """Add edges into a graphviz digraph."""
    # Setup
    metadata = MagicMock(spec_set=Metadata)

    metadata.get_tables.return_value = ['demo', 'other']
    metadata.get_parents.side_effect = [set(['other']), set()]

    metadata.get_foreign_keys.return_value = ['fk']
    metadata.get_primary_key.return_value = 'pk'

    plot = Mock()

    # Run
    visualization._add_edges(metadata, plot, names=True, details=True)

    # Asserts
    expected_edge_label = '   {}.{} > {}.{}'.format('demo', 'fk', 'other', 'pk')

    metadata.get_tables.assert_called_once_with()
    metadata.get_foreign_keys.assert_called_once_with('other', 'demo')
    metadata.get_primary_key.assert_called_once_with('other')
    assert metadata.get_parents.call_args_list == [call('demo'), call('other')]

    plot.edge.assert_called_once_with(
        'other',
        'demo',
        label=expected_edge_label,
        arrowhead='oinv'
    )


@patch('sdv.metadata.visualization.graphviz.Digraph', spec_set=graphviz.Digraph)
@patch('sdv.metadata.visualization._add_nodes', spec_set=visualization._add_nodes)
@patch('sdv.metadata.visualization._add_edges', spec_set=visualization._add_edges)
def test_visualize(add_nodes_mock, add_edges_mock, digraph_mock):
    """Metadata visualize digraph"""
    # Setup
    # plot = Mock(spec_set=graphviz.Digraph)
    # graphviz_mock.Digraph.return_value = plot

    metadata = MagicMock(spec_set=Metadata)

    # Run
    visualization.visualize(metadata, path='output.png')

    # Asserts
    digraph = digraph_mock.return_value
    add_nodes_mock.assert_called_once_with(metadata, digraph, True, True)
    add_edges_mock.assert_called_once_with(metadata, digraph, True, True)

    digraph.render.assert_called_once_with(filename='output', cleanup=True, format='png')



"""Base Class for timeseries models."""

import copy
import logging
import uuid

import cloudpickle
import pandas as pd
import rdt

from sdv.metadata import Table
from sdv.tabular.copulas import GaussianCopula
from sdv.utils import get_package_versions, throw_version_mismatch_warning

LOGGER = logging.getLogger(__name__)


class BaseTimeseriesModel:
    """Base class for timeseries models.

    Args:
        field_names (list[str]):
            List of names of the fields that need to be modeled
            and included in the generated output data. Any additional
            fields found in the data will be ignored and will not be
            included in the generated output.
            If ``None``, all the fields found in the data are used.
        field_types (dict[str, dict]):
            Dictinary specifying the data types and subtypes
            of the fields that will be modeled. Field types and subtypes
            combinations must be compatible with the SDV Metadata Schema.
        anonymize_fields (dict[str, str]):
            Dict specifying which fields to anonymize and what faker
            category they belong to.
        primary_key (str):
            Name of the field which is the primary key of the table.
        entity_columns (list[str]):
            Names of the columns which identify different time series
            sequences. These will be used to group the data in separated
            training examples.
        context_columns (list[str]):
            The columns in the dataframe which are constant within each
            group/entity. These columns will be provided at sampling time
            (i.e. the samples will be conditioned on the context variables).
        segment_size (int, pd.Timedelta or str):
            If specified, cut each training sequence in several segments of
            the indicated size. The size can either can passed as an integer
            value, which will interpreted as the number of data points to
            put on each segment, or as a pd.Timedelta (or equivalent str
            representation), which will be interpreted as the segment length
            in time. Timedelta segment sizes can only be used with sequence
            indexes of type datetime.
        sequence_index (str):
            Name of the column that acts as the order index of each
            sequence. The sequence index column can be of any type that can
            be sorted, such as integer values or datetimes.
        context_model (str or sdv.tabular.BaseTabularModel):
            Model to use to sample the context rows. It can be passed as a
            a string, which must be one of the following:

            * `gaussian_copula` (default): Use a GaussianCopula model.

            Alternatively, a preconfigured Tabular model instance can be
            passed.

        table_metadata (dict or metadata.Table):
            Table metadata instance or dict representation.
            If given alongside any other metadata-related arguments, an
            exception will be raised.
            If not given at all, it will be built using the other
            arguments or learned from the data.
    """

    _DTYPE_TRANSFORMERS = {
        'i': None,
        'f': None,
        'M': rdt.transformers.UnixTimestampEncoder(),
        'b': None,
        'O': None,
    }
    _CONTEXT_MODELS = {
        'gaussian_copula': (GaussianCopula, {'categorical_transformer': 'FrequencyEncoder_noised'})
    }

    _metadata = None

    def __init__(self, field_names=None, field_types=None, anonymize_fields=None,
                 primary_key=None, entity_columns=None, context_columns=None,
                 sequence_index=None, segment_size=None, context_model=None,
                 table_metadata=None):
        if table_metadata is None:
            self._metadata = Table(
                field_names=field_names,
                primary_key=primary_key,
                field_types=field_types,
                anonymize_fields=anonymize_fields,
                dtype_transformers=self._DTYPE_TRANSFORMERS,
                sequence_index=sequence_index,
                entity_columns=entity_columns,
                context_columns=context_columns,
            )
            self._metadata_fitted = False
        else:
            null_args = (
                field_names,
                primary_key,
                field_types,
                anonymize_fields,
                sequence_index,
                entity_columns,
                context_columns
            )
            for arg in null_args:
                if arg:
                    raise ValueError(
                        'If table_metadata is given {} must be None'.format(arg.__name__))

            if isinstance(table_metadata, dict):
                table_metadata = Table.from_dict(
                    table_metadata,
                    dtype_transformers=self._DTYPE_TRANSFORMERS,
                )

            self._metadata = table_metadata
            self._metadata_fitted = table_metadata.fitted

        # Validate arguments
        if segment_size is not None and not isinstance(segment_size, int):
            if sequence_index is None:
                raise TypeError(
                    '`segment_size` must be of type `int` if '
                    'no `sequence_index` is given.'
                )

            segment_size = pd.to_timedelta(segment_size)

        self._context_columns = self._metadata._context_columns
        self._entity_columns = self._metadata._entity_columns
        self._sequence_index = self._metadata._sequence_index
        self._segment_size = segment_size

        context_model = context_model or 'gaussian_copula'
        if isinstance(context_model, str):
            context_model = self._CONTEXT_MODELS[context_model]

        self._context_model_template = context_model

    def _fit(self, timeseries_data):
        raise NotImplementedError()

    def _fit_context_model(self, transformed):
        template = self._context_model_template
        default_kwargs = {
            'primary_key': self._entity_columns,
            'field_types': {
                name: meta
                for name, meta in self._metadata.get_fields().items()
                if name in self._entity_columns
            }
        }
        if isinstance(template, tuple):
            context_model_class, context_model_kwargs = copy.deepcopy(template)
            if 'primary_key' not in context_model_kwargs:
                context_model_kwargs['primary_key'] = self._entity_columns
                for keyword, argument in default_kwargs.items():
                    if keyword not in context_model_kwargs:
                        context_model_kwargs[keyword] = argument

            self._context_model = context_model_class(**context_model_kwargs)
        elif isinstance(template, type):
            self._context_model = template(**default_kwargs)
        else:
            self._context_model = copy.deepcopy(template)

        LOGGER.debug('Fitting context model %s', self._context_model.__class__.__name__)
        if self._context_columns:
            context = transformed[self._entity_columns + self._context_columns]
        else:
            context = transformed[self._entity_columns].copy()
            # Add constant column to allow modeling
            context[str(uuid.uuid4())] = 0

        context = context.groupby(self._entity_columns).first().reset_index()
        self._context_model.fit(context)

    def fit(self, timeseries_data):
        """Fit this model to the data.

        Args:
            timseries_data (pandas.DataFrame):
                pandas.DataFrame containing both the sequences,
                the entity columns and the context columns.
        """
        LOGGER.debug('Fitting %s to table %s; shape: %s', self.__class__.__name__,
                     self._metadata.name, timeseries_data.shape)
        if not self._metadata_fitted:
            self._metadata.fit(timeseries_data)

        LOGGER.debug('Transforming table %s; shape: %s',
                     self._metadata.name, timeseries_data.shape)
        transformed = self._metadata.transform(timeseries_data)

        for column in self._entity_columns:
            transformed[column] = timeseries_data[column]

        if self._entity_columns:
            self._fit_context_model(transformed)

        LOGGER.debug('Fitting %s model to table %s', self.__class__.__name__, self._metadata.name)
        self._fit(transformed)

    def get_metadata(self):
        """Get metadata about the table.

        This will return an ``sdv.metadata.Table`` object containing
        the information about the data that this model has learned.

        This Table metadata will contain some common information,
        such as field names and data types, as well as additional
        information that each Sub-class might add, such as the
        observed data field distributions and their parameters.

        Returns:
            sdv.metadata.Table:
                Table metadata.
        """
        return self._metadata

    def _sample(self, context=None, sequence_length=None):
        raise NotImplementedError()

    def sample(self, num_sequences=None, context=None, sequence_length=None):
        """Sample new sequences.

        Args:
            num_sequences (int):
                Number of sequences to sample. If context is
                passed, this is ignored. If not given, the
                same number of sequences as in the original
                timeseries_data is sampled.
            context (pandas.DataFrame):
                Context values to use when generating the sequences.
                If not passed, the context values will be sampled
                using the specified tabular model.
            sequence_length (int):
                If passed, sample sequences of this length. If not
                given, the sequence length will be sampled from
                the model.

        Returns:
            pandas.DataFrame:
                Table containing the sampled sequences in the same
                format as that he training data had.
        """
        if not self._entity_columns:
            if context is not None:
                raise TypeError('If there are no entity_columns, context must be None')

            context = pd.DataFrame(index=range(num_sequences or 1))
        elif context is None:
            context = self._context_model._sample_with_progress_bar(
                num_sequences, output_file_path='disable', show_progress_bar=False)
            for column in self._entity_columns or []:
                if column not in context:
                    context[column] = range(len(context))

        sampled = self._sample(context, sequence_length)
        return self._metadata.reverse_transform(sampled)

    def save(self, path):
        """Save this model instance to the given path using cloudpickle.

        Args:
            path (str):
                Path where the SDV instance will be serialized.
        """
        self._package_versions = get_package_versions(getattr(self, '_model', None))

        with open(path, 'wb') as output:
            cloudpickle.dump(self, output)

    @classmethod
    def load(cls, path):
        """Load a TabularModel instance from a given path.

        Args:
            path (str):
                Path from which to load the instance.

        Returns:
            TabularModel:
                The loaded tabular model.
        """
        with open(path, 'rb') as f:
            model = cloudpickle.load(f)
            throw_version_mismatch_warning(getattr(model, '_package_versions', None))

            return model



"""Base class for tabular model presets."""

import logging
import sys
import warnings

import cloudpickle
import numpy as np
import rdt

from sdv.metadata import Table
from sdv.tabular import GaussianCopula
from sdv.utils import get_package_versions, throw_version_mismatch_warning

LOGGER = logging.getLogger(__name__)

FAST_ML_PRESET = 'FAST_ML'
PRESETS = {
    FAST_ML_PRESET: 'Use this preset to minimize the time needed to create a synthetic data model.'
}


class TabularPreset():
    """Class for all tabular model presets.

    Args:
        name (str):
            The preset to use.
        metadata (dict or metadata.Table):
            Table metadata instance or dict representation.
        constraints (list[Constraint, dict]):
            List of Constraint objects or dicts.
    """

    _model = None
    _null_percentages = None
    _null_column = False
    _default_model = GaussianCopula

    def __init__(self, name=None, metadata=None, constraints=None):
        if name is None:
            raise ValueError('You must provide the name of a preset using the `name` '
                             'parameter. Use `TabularPreset.list_available_presets()` to browse '
                             'through the options.')
        if name not in PRESETS:
            raise ValueError(f'`name` must be one of {PRESETS}.')

        self.name = name

        if metadata is None:
            warnings.warn('No metadata provided. Metadata will be automatically '
                          'detected from your data. This process may not be accurate. '
                          'We recommend writing metadata to ensure correct data handling.')

        if metadata is not None and isinstance(metadata, Table):
            metadata = metadata.to_dict()

        if metadata is not None and constraints is not None:
            metadata['constraints'] = []
            for constraint in constraints:
                metadata['constraints'].append(constraint.to_dict())

            constraints = None

        if name == FAST_ML_PRESET:
            self._model = GaussianCopula(
                table_metadata=metadata,
                constraints=constraints,
                categorical_transformer='FrequencyEncoder_noised',
                default_distribution='gaussian',
                learn_rounding_scheme=False,
            )

            # Decide if transformers should model the null column or not.
            self._null_column = constraints is not None
            if metadata is not None:
                self._null_column = len(metadata.get('constraints', [])) > 0

            # If transformers should model the null column, pass None to let each transformer
            # decide if it's necessary or not.
            transformer_model_missing_values = bool(self._null_column)

            dtype_transformers = {
                'i': rdt.transformers.FloatFormatter(
                    missing_value_replacement='mean',
                    model_missing_values=transformer_model_missing_values,
                    enforce_min_max_values=True,
                ),
                'f': rdt.transformers.FloatFormatter(
                    missing_value_replacement='mean',
                    model_missing_values=transformer_model_missing_values,
                    enforce_min_max_values=True,
                ),
                'O': rdt.transformers.FrequencyEncoder(add_noise=True),
                'b': rdt.transformers.BinaryEncoder(
                    missing_value_replacement=-1 if self._null_column else 'mode',
                    model_missing_values=transformer_model_missing_values,
                ),
                'M': rdt.transformers.UnixTimestampEncoder(
                    missing_value_replacement='mean' if self._null_column else 'mode',
                    model_missing_values=transformer_model_missing_values,
                ),
            }
            self._model._metadata._dtype_transformers.update(dtype_transformers)

    def fit(self, data):
        """Fit this model to the data.

        Args:
            data (pandas.DataFrame):
                Data to fit the model to.
        """
        if not self._null_column:
            self._null_percentages = {}

            for column, column_data in data.iteritems():
                num_nulls = column_data.isna().sum()
                if num_nulls > 0:
                    # Store null percentage for future reference.
                    self._null_percentages[column] = num_nulls / len(column_data)

        self._model.fit(data)

    def _postprocess_sampled(self, sampled):
        """Postprocess the sampled data.

        Add null values back based on null percentages captured in the fitting process.

        Args:
            sampled (pandas.DataFrame):
                The sampled data to postprocess.

        Returns:
            pandas.DataFrame
        """
        if self._null_percentages:
            for column, percentage in self._null_percentages.items():
                sampled[column] = sampled[column].mask(
                    np.random.random((len(sampled), )) < percentage)

        return sampled

    def sample(self, num_rows, randomize_samples=True, max_tries_per_batch=100,
               batch_size=None, output_file_path=None, conditions=None):
        """Sample rows from this table.

        Args:
            num_rows (int):
                Number of rows to sample. This parameter is required.
            randomize_samples (bool):
                Whether or not to use a fixed seed when sampling. Defaults
                to True.
            max_tries_per_batch (int):
                Number of times to try sampling discarded rows. Defaults to 100.
            batch_size (int or None):
                The batch size to sample. Defaults to `num_rows`, if None.
            output_file_path (str or None):
                The file to periodically write sampled rows to. If None, does not
                write rows anywhere.
            conditions:
                Deprecated argument. Use the `sample_conditions` method with
                `sdv.sampling.Condition` objects instead.

        Returns:
            pandas.DataFrame:
                Sampled data.
        """
        sampled = self._model.sample(
            num_rows, randomize_samples, max_tries_per_batch,
            batch_size, output_file_path, conditions
        )

        return self._postprocess_sampled(sampled)

    def sample_conditions(self, conditions, max_tries_per_batch=100, batch_size=None,
                          randomize_samples=True, output_file_path=None):
        """Sample rows from this table with the given conditions.

        Args:
            conditions (list[sdv.sampling.Condition]):
                A list of sdv.sampling.Condition objects, which specify the column
                values in a condition, along with the number of rows for that
                condition.
            max_tries_per_batch (int):
                Number of times to try sampling discarded rows. Defaults to 100.
            batch_size (int):
                The batch size to use per attempt at sampling. Defaults to 10 times
                the number of rows.
            randomize_samples (bool):
                Whether or not to use a fixed seed when sampling. Defaults
                to True.
            output_file_path (str or None):
                The file to periodically write sampled rows to. Defaults to
                a temporary file, if None.

        Returns:
            pandas.DataFrame:
                Sampled data.
        """
        sampled = self._model.sample_conditions(
            conditions, max_tries_per_batch, batch_size, randomize_samples, output_file_path)

        return self._postprocess_sampled(sampled)

    def sample_remaining_columns(self, known_columns, max_tries_per_batch=100, batch_size=None,
                                 randomize_samples=True, output_file_path=None):
        """Sample rows from this table.

        Args:
            known_columns (pandas.DataFrame):
                A pandas.DataFrame with the columns that are already known. The output
                is a DataFrame such that each row in the output is sampled
                conditionally on the corresponding row in the input.
            max_tries_per_batch (int):
                Number of times to try sampling discarded rows. Defaults to 100.
            batch_size (int):
                The batch size to use per attempt at sampling. Defaults to 10 times
                the number of rows.
            randomize_samples (bool):
                Whether or not to use a fixed seed when sampling. Defaults
                to True.
            output_file_path (str or None):
                The file to periodically write sampled rows to. Defaults to
                a temporary file, if None.

        Returns:
            pandas.DataFrame:
                Sampled data.
        """
        sampled = self._model.sample_remaining_columns(
            known_columns, max_tries_per_batch, batch_size, randomize_samples, output_file_path)

        return self._postprocess_sampled(sampled)

    def save(self, path):
        """Save this model instance to the given path using cloudpickle.

        Args:
            path (str):
                Path where the SDV instance will be serialized.
        """
        self._package_versions = get_package_versions(getattr(self, '_model', None))

        with open(path, 'wb') as output:
            cloudpickle.dump(self, output)

    @classmethod
    def load(cls, path):
        """Load a TabularModel instance from a given path.

        Args:
            path (str):
                Path from which to load the instance.

        Returns:
            TabularModel:
                The loaded tabular model.
        """
        with open(path, 'rb') as f:
            model = cloudpickle.load(f)
            throw_version_mismatch_warning(getattr(model, '_package_versions', None))

            return model

    @classmethod
    def list_available_presets(cls, out=sys.stdout):
        """List the available presets and their descriptions."""
        out.write(f'Available presets:\n{PRESETS}\n\n'
                  'Supply the desired preset using the `name` parameter.\n\n'
                  'Have any requests for custom presets? Contact the SDV team to learn '
                  'more an SDV Premium license.\n')

    def __repr__(self):
        """Represent tabular preset instance as text.

        Returns:
            str
        """
        return f'TabularPreset(name={self.name})'



import io
from unittest.mock import MagicMock, Mock, mock_open, patch

import cloudpickle
import numpy as np
import pandas as pd
import pytest

from sdv.lite import TabularPreset
from sdv.metadata import Table
from sdv.tabular import GaussianCopula
from tests.utils import DataFrameMatcher


class TestTabularPreset:

    def test___init__missing_name(self):
        """Test the ``TabularPreset.__init__`` method with no parameters.

        Side Effects:
        - ValueError should be thrown
        """
        # Run and Assert
        with pytest.raises(
            ValueError,
            match=('You must provide the name of a preset using the `name` parameter. '
                   r'Use `TabularPreset.list_available_presets\(\)` to browse through '
                   'the options.')):
            TabularPreset()

    def test___init__invalid_name(self):
        """Test the ``TabularPreset.__init__`` method with an invalid arg value.

        Input:
        - name = invalid parameter

        Side Effects:
        - ValueError should be thrown
        """
        # Run and Assert
        with pytest.raises(ValueError, match=r'`name` must be one of *'):
            TabularPreset(name='invalid')

    @patch('sdv.lite.tabular.rdt.transformers')
    @patch('sdv.lite.tabular.GaussianCopula', spec_set=GaussianCopula)
    def test__init__speed_passes_correct_parameters(self, gaussian_copula_mock, transformers_mock):
        """Tests the ``TabularPreset.__init__`` method with the speed preset.

        The method should pass the parameters to the ``GaussianCopula`` class.

        Input:
        - name of the speed preset
        Side Effects:
        - GaussianCopula should receive the correct parameters
        """
        # Run
        TabularPreset(name='FAST_ML')

        # Assert
        gaussian_copula_mock.assert_called_once_with(
            table_metadata=None,
            constraints=None,
            categorical_transformer='FrequencyEncoder_noised',
            default_distribution='gaussian',
            learn_rounding_scheme=False,
        )
        metadata = gaussian_copula_mock.return_value._metadata
        metadata._dtype_transformers.update.assert_called_once_with({
            'i': transformers_mock.FloatFormatter(
                missing_value_replacement=None,
                model_missing_values=False,
                enforce_min_max_values=True,
            ),
            'f': transformers_mock.FloatFormatter(
                missing_value_replacement=None,
                model_missing_values=False,
                enforce_min_max_values=True,
            ),
            'O': transformers_mock.FrequencyEncoder(add_noise=True),
            'b': transformers_mock.BinaryEncoder(
                missing_value_replacement=None,
                model_missing_values=False
            ),
            'M': transformers_mock.UnixTimestampEncoder(
                missing_value_replacement=None,
                model_missing_values=False
            )
        })

    @patch('sdv.lite.tabular.GaussianCopula', spec_set=GaussianCopula)
    def test__init__with_metadata(self, gaussian_copula_mock):
        """Tests the ``TabularPreset.__init__`` method with the speed preset.

        The method should pass the parameters to the ``GaussianCopula`` class.

        Input:
        - name of the speed preset
        Side Effects:
        - GaussianCopula should receive the correct parameters
        """
        # Setup
        metadata = MagicMock(spec_set=Table)

        # Run
        TabularPreset(name='FAST_ML', metadata=metadata)

        # Assert
        gaussian_copula_mock.assert_called_once_with(
            table_metadata=metadata.to_dict(),
            constraints=None,
            categorical_transformer='FrequencyEncoder_noised',
            default_distribution='gaussian',
            learn_rounding_scheme=False,
        )

    @patch('sdv.lite.tabular.rdt.transformers')
    @patch('sdv.lite.tabular.GaussianCopula', spec_set=GaussianCopula)
    def test__init__with_constraints(self, gaussian_copula_mock, transformers_mock):
        """Tests the ``TabularPreset.__init__`` method with constraints.

        The constraints should be added to the metadata.

        Input:
        - constraints

        Side Effects:
        - GaussianCopula should receive args, including the constraints.
        """
        # Setup
        constraint = Mock()

        # Run
        preset = TabularPreset(name='FAST_ML', metadata=None, constraints=[constraint])

        # Assert
        gaussian_copula_mock.assert_called_once_with(
            table_metadata=None,
            constraints=[constraint],
            categorical_transformer='FrequencyEncoder_noised',
            default_distribution='gaussian',
            learn_rounding_scheme=False,
        )
        metadata = gaussian_copula_mock.return_value._metadata
        metadata._dtype_transformers.update.assert_called_once_with({
            'i': transformers_mock.FloatFormatter(
                missing_value_replacement='mean',
                model_missing_values=False,
                enforce_min_max_values=True,
            ),
            'f': transformers_mock.FloatFormatter(
                missing_value_replacement='mean',
                model_missing_values=False,
                enforce_min_max_values=True,
            ),
            'O': transformers_mock.FrequencyEncoder(add_noise=True),
            'b': transformers_mock.BinaryEncoder(
                missing_value_replacement=-1,
                model_missing_values=False
            ),
            'M': transformers_mock.UnixTimestampEncoder(
                missing_value_replacement='mean',
                model_missing_values=False
            )
        })
        assert preset._null_column is True

    @patch('sdv.lite.tabular.GaussianCopula', spec_set=GaussianCopula)
    def test__init__with_constraints_and_metadata(self, gaussian_copula_mock):
        """Tests the ``TabularPreset.__init__`` method with constraints and metadata.

        The constraints should be added to the metadata.

        Input:
        - constraints
        - metadata

        Side Effects:
        - GaussianCopula should receive metadata with the constraints added.
        """
        # Setup
        metadata = {'name': 'test_table', 'fields': []}
        constraint = Mock()

        # Run
        preset = TabularPreset(name='FAST_ML', metadata=metadata, constraints=[constraint])

        # Assert
        expected_metadata = metadata.copy()
        expected_metadata['constraints'] = [constraint.to_dict.return_value]

        gaussian_copula_mock.assert_called_once_with(
            table_metadata=expected_metadata,
            constraints=None,
            categorical_transformer='FrequencyEncoder_noised',
            default_distribution='gaussian',
            learn_rounding_scheme=False,
        )
        metadata = gaussian_copula_mock.return_value._metadata
        assert metadata._dtype_transformers.update.call_count == 1
        assert preset._null_column is True

    def test_fit(self):
        """Test the ``TabularPreset.fit`` method.

        Expect that the model's fit method is called with the expected args.

        Input:
        - fit data

        Side Effects:
        - The model's ``fit`` method is called with the same data.
        """
        # Setup
        metadata = Mock()
        metadata.to_dict.return_value = {'fields': {}}
        model = Mock()
        model._metadata = metadata
        preset = Mock()
        preset._model = model
        preset._null_percentages = None

        # Run
        TabularPreset.fit(preset, pd.DataFrame())

        # Assert
        model.fit.assert_called_once_with(DataFrameMatcher(pd.DataFrame()))
        assert preset._null_percentages is None

    def test_fit_null_column_True(self):
        """Test the ``TabularPreset.fit`` method with modeling null columns.

        Expect that the model's fit method is called with the expected args when
        ``_null_column`` is set to ``True``.

        Setup:
        - _null_column is True

        Input:
        - fit data

        Side Effects:
        - The model's ``fit`` method is called with the same data.
        - ``_null_percentages`` is ``None``
        """
        # Setup
        metadata = Mock()
        metadata.to_dict.return_value = {'fields': {}}
        model = Mock()
        model._metadata = metadata
        preset = Mock()
        preset._model = model
        preset._null_column = True
        preset._null_percentages = None

        # Run
        TabularPreset.fit(preset, pd.DataFrame())

        # Assert
        model.fit.assert_called_once_with(DataFrameMatcher(pd.DataFrame()))
        assert preset._null_percentages is None

    def test_fit_with_null_values(self):
        """Test the ``TabularPreset.fit`` method with null values.

        Expect that the model's fit method is called with the expected args, and that
        the null percentage is calculated correctly.

        Input:
        - fit data

        Side Effects:
        - The model's ``fit`` method is called with the same data.
        """
        # Setup
        metadata = Mock()
        metadata.to_dict.return_value = {'fields': {'a': {}}}
        model = Mock()
        model._metadata = metadata
        preset = Mock()
        preset._model = model
        preset._null_column = False
        preset._null_percentages = None

        data = {'a': [1, 2, np.nan]}

        # Run
        TabularPreset.fit(preset, pd.DataFrame(data))

        # Assert
        model.fit.assert_called_once_with(DataFrameMatcher(pd.DataFrame(data)))
        assert preset._null_percentages == {'a': 1.0 / 3}

    def test__postprocess_sampled_with_null_values(self):
        """Test the ``TabularPreset._postprocess_sampled`` method with null percentages.

        Expect that null values are inserted back into the sampled data.

        Setup:
        - _null_percentages has a valid entry

        Input:
        - sampled data

        Output:
        - sampled data with nulls that represents the expected null percentages.
        """
        # Setup
        sampled = pd.DataFrame({'a': [1, 2, 3, 4, 5]})
        preset = Mock()
        # Convoluted example - 100% percent chance of nulls to make test deterministic.
        preset._null_percentages = {'a': 1}

        # Run
        sampled_with_nulls = TabularPreset._postprocess_sampled(preset, sampled)

        # Assert
        assert sampled_with_nulls['a'].isna().sum() == 5

    def test_sample(self):
        """Test the ``TabularPreset.sample`` method.

        Expect that the model's sample method is called with the expected args.

        Input:
        - num_rows=5

        Side Effects:
        - The model's ``sample`` method is called with the same data.
        """
        # Setup
        model = Mock()
        preset = Mock()
        preset._model = model
        preset._null_percentages = None

        # Run
        TabularPreset.sample(preset, 5)

        # Assert
        model.sample.assert_called_once_with(5, True, 100, None, None, None)

    def test_sample_conditions(self):
        """Test the ``TabularPreset.sample_conditions`` method.

        Expect that the model's sample_conditions method is called with the expected args.

        Input:
        - num_rows=5

        Side Effects:
        - The model's ``sample_conditions`` method is called with the same data.
        """
        # Setup
        model = Mock()
        preset = Mock()
        preset._model = model
        preset._null_percentages = None
        conditions = [Mock()]

        # Run
        TabularPreset.sample_conditions(preset, conditions)

        # Assert
        model.sample_conditions.assert_called_once_with(conditions, 100, None, True, None)

    def test_sample_conditions_with_max_tries(self):
        """Test the ``TabularPreset.sample_conditions`` method with max tries.

        Expect that the model's sample_conditions method is called with the expected args.

        Input:
        - num_rows=5
        - max_retries=2
        """
        # Setup
        model = MagicMock(spec=GaussianCopula)
        preset = Mock()
        preset._model = model
        preset._null_percentages = None
        conditions = [Mock()]

        # Run
        TabularPreset.sample_conditions(preset, conditions, max_tries_per_batch=2, batch_size=5)

        # Assert
        model.sample_conditions.assert_called_once_with(conditions, 2, 5, True, None)

    def test_sample_remaining_columns(self):
        """Test the ``TabularPreset.sample_remaining_columns`` method.

        Expect that the model's sample_remaining_columns method is called with the expected args.

        Input:
        - num_rows=5

        Side Effects:
        - The model's ``sample_remaining_columns`` method is called with the same data.
        """
        # Setup
        model = Mock()
        preset = Mock()
        preset._model = model
        preset._null_percentages = None
        conditions = pd.DataFrame({'a': [1, 2, 3]})

        # Run
        TabularPreset.sample_remaining_columns(preset, conditions)

        # Assert
        model.sample_remaining_columns.assert_called_once_with(conditions, 100, None, True, None)

    def test_sample_remaining_columns_with_max_tries(self):
        """Test the ``TabularPreset.sample_remaining_columns`` method with max tries.

        Expect that the model's sample_remaining_columns method is called with the expected args.

        Input:
        - num_rows=5
        - max_retries=2
        """
        # Setup
        model = MagicMock(spec=GaussianCopula)
        preset = Mock()
        preset._model = model
        preset._null_percentages = None
        conditions = pd.DataFrame({'a': [1, 2, 3]})

        # Run
        TabularPreset.sample_remaining_columns(
            preset, conditions, max_tries_per_batch=2, batch_size=5)

        # Assert
        model.sample_remaining_columns.assert_called_once_with(conditions, 2, 5, True, None)

    def test_list_available_presets(self):
        """Tests the ``TabularPreset.list_available_presets`` method.

        This method should print all the available presets.

        Side Effects:
        - The available presets should be printed.
        """
        # Setup
        out = io.StringIO()
        expected = ('Available presets:\n{\'FAST_ML\': \'Use this preset to minimize the time '
                    'needed to create a synthetic data model.\'}\n\nSupply the desired '
                    'preset using the `name` parameter.\n\nHave any requests for '
                    'custom presets? Contact the SDV team to learn more an SDV Premium license.')

        # Run
        TabularPreset.list_available_presets(out)

        # Assert
        assert out.getvalue().strip() == expected

    @patch('sdv.lite.tabular.cloudpickle')
    def test_save(self, cloudpickle_mock):
        """Test the ``TabularPreset.save`` method.

        Expect that the model's save method is called with the expected args.

        Input:
        - path

        Side Effects:
        - The model's ``save`` method is called with the same argument.
        """
        # Setup
        model = Mock()
        preset = Mock()
        preset._model = model
        open_mock = mock_open(read_data=cloudpickle.dumps('test'))

        # Run
        with patch('sdv.lite.tabular.open', open_mock):
            TabularPreset.save(preset, 'test-path')

        # Assert
        open_mock.assert_called_once_with('test-path', 'wb')
        cloudpickle_mock.dump.assert_called_once_with(preset, open_mock())

    @patch('sdv.lite.tabular.cloudpickle')
    def test_load(self, cloudpickle_mock):
        """Test the ``TabularPreset.load`` method.

        Expect that the model's load method is called with the expected args.

        Input:
        - path

        Side Effects:
        - The default model's ``load`` method is called with the same argument.
        """
        # Setup
        default_model = Mock()
        TabularPreset._default_model = default_model
        open_mock = mock_open(read_data=cloudpickle.dumps('test'))

        # Run
        with patch('sdv.lite.tabular.open', open_mock):
            loaded = TabularPreset.load('test-file.pkl')

        # Assert
        open_mock.assert_called_once_with('test-file.pkl', 'rb')
        assert loaded == cloudpickle_mock.load.return_value

    def test___repr__(self):
        """Test the ``TabularPreset.__repr__`` method.

        Output:
        - Expect a string 'TabularPreset(name=<name>)'
        """
        # Setup
        instance = TabularPreset('FAST_ML')

        # Run
        res = repr(instance)

        # Assert
        assert res == 'TabularPreset(name=FAST_ML)'



import numpy as np
import pandas as pd

from sdv.demo import load_demo
from sdv.sampling import Condition
from sdv.tabular.ctgan import TVAE


def test_tvae():
    users = load_demo(metadata=False)['users']

    tvae = TVAE(
        primary_key='user_id',
        epochs=1
    )
    tvae.fit(users)

    sampled = tvae.sample(len(users))

    # test shape is right
    assert sampled.shape == users.shape

    # test user_id has been generated as an ID field
    assert list(sampled['user_id']) == list(range(0, len(users)))

    expected_metadata = {
        'fields': {
            'user_id': {
                'type': 'id',
                'subtype': 'integer',
                'transformer': 'FloatFormatter',
            },
            'country': {
                'type': 'categorical',
                'transformer': None,
            },
            'gender': {
                'type': 'categorical',
                'transformer': None,
            },
            'age': {
                'type': 'numerical',
                'subtype': 'integer',
                'transformer': 'FloatFormatter',
            }
        },
        'primary_key': 'user_id',
        'constraints': [],
        'sequence_index': None,
        'context_columns': [],
        'entity_columns': [],
        'model_kwargs': {},
        'name': None
    }
    assert tvae.get_metadata().to_dict() == expected_metadata


def test_recreate():
    data = load_demo(metadata=False)['users']

    # If distribution is non parametric, get_parameters fails
    model = TVAE(epochs=1)
    model.fit(data)
    sampled = model.sample(len(data))

    assert sampled.shape == data.shape
    assert (sampled.dtypes == data.dtypes).all()
    assert (sampled.notnull().sum(axis=1) != 0).all()

    # Metadata
    model_meta = TVAE(epochs=1, table_metadata=model.get_metadata())
    model_meta.fit(data)
    sampled = model_meta.sample(len(data))

    assert sampled.shape == data.shape
    assert (sampled.dtypes == data.dtypes).all()
    assert (sampled.notnull().sum(axis=1) != 0).all()

    # Metadata dict
    model_meta_dict = TVAE(epochs=1, table_metadata=model.get_metadata().to_dict())
    model_meta_dict.fit(data)
    sampled = model_meta_dict.sample(len(data))

    assert sampled.shape == data.shape
    assert (sampled.dtypes == data.dtypes).all()
    assert (sampled.notnull().sum(axis=1) != 0).all()


def test_conditional_sampling_dict():
    data = pd.DataFrame({
        'column1': [1.0, 0.5, 2.5] * 10,
        'column2': ['a', 'b', 'c'] * 10
    })

    model = TVAE(epochs=1)
    model.fit(data)
    conditions = [Condition({
        'column2': 'b'
    }, num_rows=30)]
    sampled = model.sample_conditions(conditions=conditions)

    assert sampled.shape == data.shape
    assert set(sampled['column2'].unique()) == set(['b'])


def test_conditional_sampling_dataframe():
    data = pd.DataFrame({
        'column1': [1.0, 0.5, 2.5] * 10,
        'column2': ['a', 'b', 'c'] * 10
    })

    model = TVAE(epochs=1)
    model.fit(data)
    conditions = pd.DataFrame({
        'column2': ['b', 'b', 'b', 'c', 'c']
    })
    sampled = model.sample_remaining_columns(conditions)

    assert sampled.shape[0] == len(conditions['column2'])
    assert (sampled['column2'] == np.array(['b', 'b', 'b', 'c', 'c'])).all()


def test_conditional_sampling_two_conditions():
    data = pd.DataFrame({
        'column1': [1.0, 0.5, 2.5] * 10,
        'column2': ['a', 'b', 'c'] * 10,
        'column3': ['d', 'e', 'f'] * 10
    })

    model = TVAE(epochs=1)
    model.fit(data)
    conditions = [Condition({
        'column2': 'b',
        'column3': 'f'
    }, num_rows=5)]
    samples = model.sample_conditions(conditions=conditions, max_tries_per_batch=200)
    assert list(samples.column2) == ['b'] * 5
    assert list(samples.column3) == ['f'] * 5


def test_conditional_sampling_numerical():
    data = pd.DataFrame({
        'column1': [1.0, 0.5, 2.5] * 10,
        'column2': ['a', 'b', 'c'] * 10,
        'column3': ['d', 'e', 'f'] * 10
    })

    model = TVAE(epochs=1)
    model.fit(data)
    conditions = [Condition({
        'column1': 1.0,
    }, num_rows=5)]
    sampled = model.sample_conditions(conditions=conditions)

    assert list(sampled.column1) == [1.0] * 5



"""Tests for the sdv.sampling.tabular module."""
from sdv.sampling.tabular import Condition


class TestCondition():

    def test___init__(self):
        """Test ```Condition.__init__`` method.

        Expect that `column_values` and `num_rows` are defined correctly.

        Input:
            - column_values
            - num_rows
        """
        # Setup
        column_values = {'a': 1, 'b': 2}
        num_rows = 5

        # Run
        condition = Condition(column_values=column_values, num_rows=num_rows)

        # Assert
        assert condition.column_values == column_values
        assert condition.num_rows == num_rows

    def test_get_column_values(self):
        """Test ```Condition.get_column_values`` method.

        Expect that the correct `column_values` value is returned.

        Input:
            - column_values
        """
        # Setup
        column_values = {'a': 1, 'b': 2}
        condition = Condition(column_values=column_values)

        # Run
        condition_column_values = condition.get_column_values()

        # Assert
        assert condition_column_values == column_values

    def test_get_num_rows_default(self):
        """Test ```Condition.get_num_rows`` method.

        Expect that the default `num_rows` value is returned, and
        that the default value is 1.

        Input:
            - column_values
        """
        # Setup
        column_values = {'a': 1, 'b': 2}
        condition = Condition(column_values=column_values)

        # Run
        default_num_rows = condition.get_num_rows()

        # Assert
        assert default_num_rows == 1

    def test_get_num_rows(self):
        """Test ```Condition.get_num_rows`` method.

        Expect that the correct `num_rows` value is returned.

        Input:
            - column_values
            - num_rows
        """
        # Setup
        column_values = {'a': 1, 'b': 2}
        num_rows = 100
        condition = Condition(column_values=column_values, num_rows=num_rows)

        # Run
        condition_num_rows = condition.get_num_rows()

        # Assert
        assert condition_num_rows == num_rows



import re

import numpy as np
import pandas as pd

from sdv.metadata import Table


def test_table():
    """Test Table with pii and non-pii columns."""
    num_values = 100

    metadata_dict = {
        'fields': {
            'years_employed': {
                'type': 'numerical',
                'subtype': 'integer',
            },
            'ssn': {
                'type': 'categorical',
                'pii': True,
                'pii_category': 'ssn',
            },
            'company_US': {
                'type': 'categorical',
                'pii': True,
                'pii_category': 'company',
                'pii_locales': ['en_US'],
            },
            'company_US_CN': {
                'type': 'categorical',
                'pii': True,
                'pii_category': 'company',
                'pii_locales': ['en_US', 'zh_CN'],
            },
        },
    }
    data = pd.DataFrame({
        'years_employed': np.random.choice(20, num_values),
        'ssn': [str(i) for i in range(num_values)],
        'company_US': [str(i) for i in range(num_values)],
        'company_US_CN': [str(i) for i in range(num_values)],
    })
    metadata = Table.from_dict(metadata_dict)

    metadata.fit(data)

    transformed = metadata.transform(data)
    assert transformed.dtypes.isin([np.dtype('int32'), np.dtype('int64'), np.dtype('float')]).all()

    reverse_transformed = metadata.reverse_transform(transformed)

    sampled_years_employed = reverse_transformed['years_employed']
    assert sampled_years_employed.dtype == 'int'
    assert ((sampled_years_employed >= 0) & (sampled_years_employed < 20)).all()

    sampled_ssn = reverse_transformed['ssn']
    ssn_regex = re.compile(r'^\d\d\d-\d\d-\d\d\d\d$')
    assert sampled_ssn.dtype == 'object'
    assert sampled_ssn.str.match(ssn_regex).all()

    sampled_company_US = reverse_transformed['company_US']
    assert sampled_company_US.dtype == 'object'
    # Check that all companies are sampled from the `en_US` locale
    assert ((sampled_company_US > u'\u0000') & (sampled_company_US < u'\u007F')).all()

    sampled_company_US_CN = reverse_transformed['company_US_CN']
    assert sampled_company_US_CN.dtype == 'object'
    # Check that we have sampled companies from the `en_US` locale
    assert ((sampled_company_US_CN > u'\u0000') & (sampled_company_US_CN < u'\u007F')).any()
    # Check that we have sampled companies from the `zh_CH` locale
    assert ((sampled_company_US_CN > u'\u4e00') & (sampled_company_US_CN < u'\u9fff')).any()



import pandas as pd

from sdv.lite import TabularPreset


def test_sample():
    """Test sampling for the ``TabularPreset``."""
    # Setup
    data = pd.DataFrame({'a': [1, 2, 3]})

    # Run
    preset = TabularPreset(name='FAST_ML')
    preset.fit(data)
    samples = preset.sample(
        num_rows=10,
        max_tries_per_batch=20,
        batch_size=5
    )

    # Assert
    assert samples['a'].all() in [1, 2, 3]
    assert len(samples) == 10



from sdv import SDV, Metadata, load_demo
from sdv.evaluation import evaluate


def test_evaluate_tables_from_demo():
    tables = load_demo(metadata=False)

    new_meta = Metadata()
    new_meta.add_table('users', data=tables['users'], primary_key='user_id')
    new_meta.add_table('sessions', data=tables['sessions'], primary_key='session_id',
                       parent='users', foreign_key='user_id')
    transactions_fields = {
        'timestamp': {
            'type': 'datetime',
            'format': '%Y-%m-%d'
        }
    }
    new_meta.add_table('transactions', tables['transactions'],
                       fields_metadata=transactions_fields,
                       primary_key='transaction_id', parent='sessions')

    sdv = SDV()
    sdv.fit(new_meta, tables=tables)

    sampled = sdv.sample_all()

    table_scores = dict()
    for table in new_meta.get_tables():
        table_scores[table] = evaluate(
            sampled[table], tables[table], metadata=new_meta, table_name=table)

    score = evaluate(sampled, tables, metadata=new_meta)

    assert isinstance(score, float)
    assert 0 <= score <= 1



"""Tests for the sdv.utils module."""

from unittest.mock import Mock, call, patch

import pkg_resources

from sdv.tabular import GaussianCopula
from sdv.utils import get_package_versions, throw_version_mismatch_warning


@patch('sdv.utils.pkg_resources.get_distribution')
def test_get_package_versions_no_model(get_distribution_mock):
    """Test the ``get_package_versions`` method when there is no model.

    Expect that the package versions for `sdv` and `rdt` are returned.

    Setup:
        - Patch the versions of sdv and rdt
    Input:
        - model is None
    Output:
        - a dict mapping libraries to library versions.
    """
    # Setup
    expected = {
        'sdv': get_distribution_mock.return_value.version,
        'rdt': get_distribution_mock.return_value.version,
    }

    # Run
    versions = get_package_versions(model=None)

    # Assert
    assert get_distribution_mock.has_calls([call('sdv'), call('rdt')])
    assert versions == expected


@patch('sdv.utils.pkg_resources.get_distribution')
def test_get_package_versions_valid_model(get_distribution_mock):
    """Test the ``get_package_versions`` method with model.

    Expect that the package versions for `sdv` and `rdt` are returned.

    Setup:
        - Patch the versions of sdv and rdt
        - Create a model named GaussianCopula
    Input:
        - model
    Output:
        - a dict mapping libraries to library versions.
    """
    # Setup
    expected = {
        'sdv': get_distribution_mock.return_value.version,
        'rdt': get_distribution_mock.return_value.version,
        'copulas': get_distribution_mock.return_value.version,
    }
    model = Mock(spec=GaussianCopula)

    # Run
    versions = get_package_versions(model=model)

    # Assert
    assert get_distribution_mock.has_calls([call('sdv'), call('rdt'), call('copulas')])
    assert versions == expected


@patch('sdv.utils.pkg_resources.get_distribution')
def test_get_package_versions_error(get_distribution_mock):
    """Test the ``get_package_versions`` method with model.

    Expect that the package versions for `sdv` and `rdt` are returned, and that the unavailable
    packages are ignored.

    Setup:
        - Patch the versions of sdv and rdt
        - Create a model named GaussianCopula
    Input:
        - model
    Output:
        - a dict mapping libraries to library versions.
    """
    # Setup
    dist_mock = Mock()
    get_distribution_mock.side_effect = [
        dist_mock, pkg_resources.ResolutionError(), dist_mock]

    expected = {
        'sdv': dist_mock.version,
        'copulas': dist_mock.version,
    }
    model = Mock(spec=GaussianCopula)

    # Run
    versions = get_package_versions(model=model)

    # Assert
    assert get_distribution_mock.has_calls([call('sdv'), call('rdt'), call('copulas')])
    assert versions == expected


@patch('sdv.utils.pkg_resources.warnings')
@patch('sdv.utils.pkg_resources.get_distribution')
def test_throw_version_mismatch_warning_no_mismatches(get_distribution_mock, warnings_mock):
    """Test the ``generate_version_mismatch_warning`` method with no mismatched versions.

    Expect that None is returned.

    Setup:
        - Mock ``pkg_resources.get_distribution`` to return the same versions.
    Input:
        - package versions mapping
    Output:
        - None
    """
    # Setup
    dist_mock = Mock()
    get_distribution_mock.side_effect = [dist_mock, dist_mock]
    package_versions = {'sdv': dist_mock.version, 'rdt': dist_mock.version}

    # Run
    throw_version_mismatch_warning(package_versions)

    # Assert
    assert get_distribution_mock.hass_calls([call('sdv'), call('rdt')])
    assert warnings_mock.warn.call_count == 0


@patch('sdv.utils.pkg_resources.warnings')
@patch('sdv.utils.pkg_resources.get_distribution')
def test_generate_version_mismatch_warning_mismatched_version(get_distribution_mock,
                                                              warnings_mock):
    """Test the ``generate_version_mismatch_warning`` method with one mismatched version.

    Expect that the proper warning is thrown.

    Setup:
        - Mock ``pkg_resources.get_distribution`` to return the same versions, except for one.
    Input:
        - package versions mapping
    Side Effect:
        - warning
    """
    # Setup
    dist_mock = Mock()
    get_distribution_mock.side_effect = [dist_mock, dist_mock]
    package_versions = {'sdv': dist_mock.version, 'rdt': 'rdt_version'}
    expected_str = ('The libraries used to create the model have older versions than your '
                    'current setup. This may cause errors when sampling.\nrdt used version '
                    '`rdt_version`; current version is `{0}`'.format(dist_mock.version))

    # Run
    throw_version_mismatch_warning(package_versions)

    # Assert
    assert get_distribution_mock.hass_calls([call('sdv'), call('rdt')])
    assert warnings_mock.warn.called_once_with(expected_str)


@patch('sdv.utils.pkg_resources.warnings')
@patch('sdv.utils.pkg_resources.get_distribution')
def test_generate_version_mismatch_warning_mismatched_version_error(get_distribution_mock,
                                                                    warnings_mock):
    """Test the ``generate_version_mismatch_warning`` method with one mismatched version.

    Expect that if we error out when trying to get the current package version for one
    package, we treat that as a mismatch and the proper warning is thrown.

    Setup:
        - Mock ``pkg_resources.get_distribution`` to return the same versions, except for one,
          which errors out.
    Input:
        - package versions mapping
    Side Effect:
        - warning
    """
    # Setup
    dist_mock = Mock()
    get_distribution_mock.side_effect = [dist_mock, pkg_resources.ResolutionError()]
    package_versions = {'sdv': dist_mock.version, 'rdt': 'rdt_version'}
    expected_str = ('The libraries used to create the model have older versions than your '
                    'current setup. This may cause errors when sampling.\nrdt used version '
                    '`rdt_version`; current version is ``')

    # Run
    throw_version_mismatch_warning(package_versions)

    # Assert
    assert get_distribution_mock.hass_calls([call('sdv'), call('rdt')])
    assert warnings_mock.warn.called_once_with(expected_str)


@patch('sdv.utils.pkg_resources.warnings')
def test_generate_version_mismatch_warning_no_package_versions(warnings_mock):
    """Test the ``generate_version_mismatch_warning`` method with None as input.

    Expect that if the model didn't have a `_package_versions` attribute, we throw
    a generic warning.

    Input:
        - package versions is None
    Side Effect:
        - Generic warning is thrown.
    """
    # Setup
    package_versions = None
    expected_str = ('The libraries used to create the model have older versions than your '
                    'current setup. This may cause errors when sampling.')

    # Run
    throw_version_mismatch_warning(package_versions)

    # Assert
    assert warnings_mock.warn.called_once_with(expected_str)
